import json
import logging
import os
import random
from typing import Optional

from sklearn.utils import compute_class_weight
import torch
import torch.nn as nn

from tqdm import tqdm
from datetime import date, datetime
from typing import Iterable, Iterator, cast
from torch.utils.data import DataLoader, Sampler
from torch.optim import AdamW
from torch.optim.lr_scheduler import MultiStepLR
import numpy

from medcat.cdb import CDB
from medcat.vocab import Vocab
from medcat.config.config import Config, ComponentConfig
from medcat.config.config_rel_cat import ConfigRelCAT
from medcat.storage.serialisers import deserialise
from medcat.storage.serialisables import SerialisingStrategy
from medcat.components.addons.addons import AddonComponent
from medcat.components.addons.relation_extraction.base_component import (
    RelExtrBaseComponent)
from medcat.components.addons.meta_cat.ml_utils import set_all_seeds
from medcat.components.addons.relation_extraction.ml_utils import (
    load_results, load_state, save_results, save_state,
    split_list_train_test_by_class)
from medcat.components.addons.relation_extraction.rel_dataset import RelData
from medcat.tokenizing.tokenizers import BaseTokenizer, create_tokenizer
from medcat.tokenizing.tokens import MutableDocument
from medcat.utils.defaults import COMPONENTS_FOLDER


logger = logging.getLogger(__name__)


class RelCATAddon(AddonComponent):
    addon_type = 'rel_cat'
    output_key = 'relations'
    config: ConfigRelCAT

    def __init__(self, config: ConfigRelCAT,
                 rel_cat: "RelCAT"):
        self.config = config
        self._rel_cat = rel_cat

    @classmethod
    def create_new(cls, config: ConfigRelCAT, base_tokenizer: BaseTokenizer,
                   cdb: CDB) -> 'RelCATAddon':
        """Factory method to create a new MetaCATAddon instance."""
        return cls(config,
                   RelCAT(base_tokenizer, cdb, config=config, init_model=True))

    @classmethod
    def create_new_component(
            cls, cnf: ComponentConfig, tokenizer: BaseTokenizer,
            cdb: CDB, vocab: Vocab, model_load_path: Optional[str]
            ) -> 'RelCATAddon':
        if not isinstance(cnf, ConfigRelCAT):
            raise ValueError(f"Incompatible config: {cnf}")
        config = cnf
        if model_load_path is not None:
            load_path = os.path.join(model_load_path, COMPONENTS_FOLDER,
                                     cls.NAME_PREFIX + cls.addon_type)
            return cls.load_existing(config, tokenizer, cdb, load_path)
        return cls.create_new(config, tokenizer, cdb)

    @classmethod
    def load_existing(cls, cnf: ConfigRelCAT,
                      base_tokenizer: BaseTokenizer,
                      cdb: CDB,
                      load_path: str) -> 'RelCATAddon':
        """Factory method to load an existing RelCAT addon from disk."""
        rc = RelCAT.load(load_path)
        # set the correct base tokenizer and redo data paths
        rc.base_tokenizer = base_tokenizer
        rc._init_data_paths()
        return cls(cnf, rc)

    def serialise_to(self, folder_path: str) -> None:
        os.mkdir(folder_path)
        self._rel_cat.save(folder_path)

    @property
    def name(self) -> str:
        return str(self.addon_type)

    # for ManualSerialisable:

    @classmethod
    def deserialise_from(cls, folder_path: str, **init_kwargs
                         ) -> 'RelCATAddon':
        # NOTE: model load path sent by kwargs
        return cls.load_existing(
            load_path=folder_path,
            base_tokenizer=init_kwargs['tokenizer'],
            cnf=init_kwargs['cnf'],
            cdb=init_kwargs['cdb'],
        )

    def get_strategy(self) -> SerialisingStrategy:
        return SerialisingStrategy.MANUAL

    @classmethod
    def get_init_attrs(cls) -> list[str]:
        return []

    @classmethod
    def ignore_attrs(cls) -> list[str]:
        return []

    @classmethod
    def include_properties(cls) -> list[str]:
        return []

    def __call__(self, doc: MutableDocument):
        return self._rel_cat(doc)


class BalancedBatchSampler(Sampler):

    def __init__(self, dataset, classes,
                 batch_size, max_samples, max_minority):
        self.dataset = dataset
        self.classes = classes
        self.batch_size = batch_size
        self.num_classes = len(classes)
        self.indices = list(range(len(dataset)))

        self.max_minority = max_minority

        self.max_samples_per_class = max_samples

    def __len__(self):
        return (len(self.indices) + self.batch_size - 1) // self.batch_size

    def __iter__(self):
        batch_counter = 0
        indices = self.indices.copy()
        while batch_counter != self.__len__():
            batch = []

            class_counts = {c: 0 for c in self.classes}
            while len(batch) < self.batch_size:

                index = random.choice(indices)
                # Assuming label is at index 1
                label = self.dataset[index][2].numpy().tolist()[0]
                if class_counts[label] < self.max_samples_per_class[label]:
                    batch.append(index)
                    class_counts[label] += 1
                    if self.max_samples_per_class[label] > self.max_minority:
                        indices.remove(index)

            yield batch
            batch_counter += 1


class RelCAT:
    """The RelCAT class used for training 'Relation-Annotation' models, i.e.,
    annotation of relations between clinical concepts.

    Args:
        cdb (CDB): cdb, this is used when creating relation datasets.

        tokenizer (TokenizerWrapperBERT):
            The Huggingface tokenizer instance. This can be a pre-trained
            tokenzier instance from a BERT-style model. For now, only
            BERT models are supported.

        config (ConfigRelCAT):
            the configuration for RelCAT. Param descriptions available in
            ConfigRelCAT class docs.

        task (str, optional): What task is this model supposed to handle.
            Defaults to "train"
        init_model (bool, optional): loads default model. Defaults to False.

    """
    addon_type = 'rel_cat'
    output_key = 'rel_'

    def __init__(self, base_tokenizer: BaseTokenizer,
                 cdb: CDB, config: ConfigRelCAT = ConfigRelCAT(),
                 task: str = "train", init_model: bool = False):
        self.base_tokenizer = base_tokenizer
        self.component = RelExtrBaseComponent()
        self.task: str = task
        self.checkpoint_path: str = "./"

        set_all_seeds(config.general.seed)

        if init_model:
            self.component = RelExtrBaseComponent(
                config=config, task=task, init_model=True)

        self.cdb = cdb
        logging.basicConfig(
            level=self.component.relcat_config.general.log_level)
        logger.setLevel(self.component.relcat_config.general.log_level)

        self.is_cuda_available = torch.cuda.is_available()
        self.device = torch.device(
            "cuda" if self.is_cuda_available and
            self.component.relcat_config.general.device != "cpu" else "cpu")
        self._init_data_paths()

    def _init_data_paths(self):
        doc_cls = self.base_tokenizer.get_doc_class()
        doc_cls.register_addon_path('relations', def_val=[], force=True)

    def save(self, save_path: str = "./") -> None:
        self.component.save(save_path=save_path)

    @classmethod
    def load(cls, load_path: str = "./") -> "RelCAT":

        if os.path.exists(os.path.join(load_path, "cdb.dat")):
            cdb = cast(CDB, deserialise(os.path.join(load_path, "cdb.dat")))
        else:
            cdb = CDB(config=Config())
            logger.info(
                "The default CDB file name 'cdb.dat' doesn't exist in the "
                "specified path, you will need to load & set "
                "a CDB manually via rel_cat.cdb = CDB.load('path') ")

        component = RelExtrBaseComponent.load(
            pretrained_model_name_or_path=load_path)

        device = torch.device(
            "cuda" if torch.cuda.is_available() and
            component.relcat_config.general.device != "cpu" else "cpu")

        rel_cat = RelCAT(
            # NOTE: this is a throaway tokenizer just for registrations
            create_tokenizer(cdb.config.general.nlp.provider, cdb.config),
            cdb=cdb, config=component.relcat_config, task=component.task)
        rel_cat.device = device
        rel_cat.component = component

        return rel_cat

    def __call__(self, doc: MutableDocument) -> MutableDocument:
        doc = next(self.pipe(iter([doc])))
        return doc

    def _create_test_train_datasets(self, data: dict,
                                    split_sets: bool = False):
        train_data: dict = {}
        test_data: dict = {}

        if split_sets:
            rc_cnf = self.component.relcat_config
            (train_data["output_relations"],
             test_data["output_relations"]) = split_list_train_test_by_class(
                 data["output_relations"],
                 test_size=rc_cnf.train.test_size,
                 shuffle=rc_cnf.train.shuffle_data,
                 sample_limit=rc_cnf.general.limit_samples_per_class)

            test_data_label_names = [
                rec[4] for rec in test_data["output_relations"]]

            (test_data["nclasses"], test_data["labels2idx"],
             test_data["idx2label"]) = RelData.get_labels(
                test_data_label_names, self.component.relcat_config)

            for idx in range(len(test_data["output_relations"])):
                test_data["output_relations"
                          ][idx][5] = test_data["labels2idx"][
                              test_data["output_relations"][idx][4]]
        else:
            train_data["output_relations"] = data["output_relations"]

        for k, v in data.items():
            if k != "output_relations":
                train_data[k] = []
                test_data[k] = []

        train_data_label_names = [rec[4]
                                  for rec in train_data["output_relations"]]

        (train_data["nclasses"], train_data["labels2idx"],
         train_data["idx2label"]) = RelData.get_labels(
            train_data_label_names, self.component.relcat_config)

        for idx in range(len(train_data["output_relations"])):
            train_data["output_relations"
                       ][idx][5] = train_data["labels2idx"][
                           train_data["output_relations"][idx][4]]

        return train_data, test_data

    def train(self, export_data_path: str = "", train_csv_path: str = "",
              test_csv_path: str = "", checkpoint_path: str = "./"):

        if self.is_cuda_available:
            logger.info("Training on device: %s%s",
                        str(torch.cuda.get_device_name(0)), str(self.device))

        self.component.model = self.component.model.to(self.device)

        rc_cnf = self.component.relcat_config

        # resize vocab just in case more tokens have been added
        self.component.model_config.vocab_size = (
            self.component.tokenizer.get_size())

        train_rel_data = RelData(
            cdb=self.cdb, config=rc_cnf,
            tokenizer=self.component.tokenizer)
        test_rel_data = RelData(
            cdb=self.cdb, config=rc_cnf,
            tokenizer=self.component.tokenizer)

        if train_csv_path != "":
            if test_csv_path != "":
                train_rel_data.dataset, _ = self._create_test_train_datasets(
                    train_rel_data.create_base_relations_from_csv(
                        train_csv_path), split_sets=False)
                test_rel_data.dataset, _ = self._create_test_train_datasets(
                    train_rel_data.create_base_relations_from_csv(
                        test_csv_path), split_sets=False)
            else:
                (train_rel_data.dataset,
                 test_rel_data.dataset) = self._create_test_train_datasets(
                    train_rel_data.create_base_relations_from_csv(
                        train_csv_path), split_sets=True)

        elif export_data_path != "":
            export_data = {}
            with open(export_data_path) as f:
                export_data = json.load(f)
            (train_rel_data.dataset,
             test_rel_data.dataset) = self._create_test_train_datasets(
                train_rel_data.create_relations_from_export(export_data),
                split_sets=True)
        else:
            raise ValueError(
                "NO DATA HAS BEEN PROVIDED (MedCAT Trainer export "
                "JSON/CSV/spacy_DOCS)")

        train_dataset_size = len(train_rel_data)
        batch_size = (
            train_dataset_size if train_dataset_size < rc_cnf.train.batch_size
            else rc_cnf.train.batch_size)

        # to use stratified batching
        if rc_cnf.train.stratified_batching:
            sampler = BalancedBatchSampler(
                train_rel_data, [
                    i for i in
                    range(rc_cnf.train.nclasses)],
                batch_size,
                rc_cnf.train.batching_samples_per_class,
                rc_cnf.train.batching_minority_limit)

            train_dataloader = DataLoader(
                train_rel_data, num_workers=0,
                collate_fn=self.component.padding_seq,
                batch_sampler=sampler,
                pin_memory=rc_cnf.general.pin_memory)
        else:
            train_dataloader = DataLoader(
                train_rel_data, batch_size=batch_size,
                shuffle=rc_cnf.train.shuffle_data,
                num_workers=0,
                collate_fn=self.component.padding_seq,
                pin_memory=rc_cnf.general.pin_memory)

        test_dataset_size = len(test_rel_data)
        test_batch_size = (
            test_dataset_size if
            test_dataset_size < rc_cnf.train.batch_size
            else rc_cnf.train.batch_size)
        test_dataloader = DataLoader(
            test_rel_data,
            batch_size=test_batch_size,
            shuffle=rc_cnf.train.shuffle_data,
            num_workers=0,
            collate_fn=self.component.padding_seq,
            pin_memory=rc_cnf.general.pin_memory)

        if (rc_cnf.train.class_weights is not None and
                rc_cnf.train.enable_class_weights):
            criterion = nn.CrossEntropyLoss(weight=torch.FloatTensor(
                numpy.asarray(rc_cnf.train.class_weights)
                ).to(self.device))
        elif rc_cnf.train.enable_class_weights:
            all_class_lbl_ids = [
                rec[5] for rec in train_rel_data.dataset["output_relations"]]
            rc_cnf.train.class_weights = (
                compute_class_weight(class_weight="balanced",
                                     classes=numpy.unique(all_class_lbl_ids),
                                     y=all_class_lbl_ids).tolist())
            criterion = nn.CrossEntropyLoss(weight=torch.FloatTensor(
                rc_cnf.train.class_weights).to(
                    self.device))
        else:
            criterion = nn.CrossEntropyLoss()

        if self.component.optimizer is None:
            parameters = filter(lambda p: p.requires_grad,
                                self.component.model.parameters())
            self.component.optimizer = AdamW(
                parameters, lr=self.component.relcat_config.train.lr,
                weight_decay=rc_cnf.train.adam_weight_decay,
                betas=rc_cnf.train.adam_betas, eps=rc_cnf.train.adam_epsilon)

        if self.component.scheduler is None:
            self.component.scheduler = MultiStepLR(
                self.component.optimizer,
                milestones=rc_cnf.train.multistep_milestones,
                gamma=rc_cnf.train.multistep_lr_gamma)

        self.epoch, self.best_f1 = load_state(
            self.component.model, self.component.optimizer,
            self.component.scheduler, load_best=False, path=checkpoint_path,
            relcat_config=rc_cnf)

        logger.info("Starting training process...")

        losses_per_epoch, accuracy_per_epoch, f1_per_epoch = load_results(
            path=checkpoint_path)

        if train_rel_data.dataset["nclasses"] > rc_cnf.train.nclasses:
            rc_cnf.train.nclasses = train_rel_data.dataset["nclasses"]
            self.component.model.relcat_config.train.nclasses = (
                rc_cnf.train.nclasses)

        rc_cnf.general.labels2idx.update(train_rel_data.dataset["labels2idx"])
        rc_cnf.general.idx2labels = {
            int(v): k for k, v in rc_cnf.general.labels2idx.items()}

        gradient_acc_steps = (
            rc_cnf.train.gradient_acc_steps)
        max_grad_norm = rc_cnf.train.max_grad_norm

        _epochs = self.epoch + rc_cnf.train.nepochs

        for epoch in range(0, _epochs):
            epoch_losses, epoch_precision, epoch_f1 = self._train_epoch(
                epoch, gradient_acc_steps, max_grad_norm, train_dataset_size,
                train_dataloader, test_dataloader, criterion, _epochs,
                checkpoint_path)
            losses_per_epoch.extend(epoch_losses)
            accuracy_per_epoch.extend(epoch_precision)
            f1_per_epoch.extend(epoch_f1)

    def _train_epoch(self, epoch: int,
                     gradient_acc_steps: int,
                     max_grad_norm: float,
                     train_dataset_size: int,
                     train_dataloader: DataLoader,
                     test_dataloader: DataLoader,
                     criterion: nn.CrossEntropyLoss,
                     _epochs: int,
                     checkpoint_path: str) -> tuple[list, list, list]:
        rc_cnf = self.component.relcat_config
        start_time = datetime.now().time()
        total_loss = 0.0

        loss_per_batch = []
        accuracy_per_batch = []

        logger.info(
            "Total epochs on this model: %d | currently training "
            "epoch %d", _epochs, epoch)

        pbar = tqdm(total=train_dataset_size)

        for i, data in enumerate(train_dataloader, 0):
            self.component.model.train()
            self.component.model.zero_grad()

            current_batch_size = len(data[0])
            token_ids, e1_e2_start, labels, _, _ = data

            attention_mask = (
                token_ids != self.component.pad_id).float().to(self.device)

            token_type_ids = torch.zeros(
                (token_ids.shape[0], token_ids.shape[1])).long().to(
                    self.device)

            labels = labels.to(self.device)

            model_output, classification_logits = self.component.model(
                input_ids=token_ids,
                token_type_ids=token_type_ids,
                attention_mask=attention_mask,
                e1_e2_start=e1_e2_start
            )

            batch_loss = criterion(
                classification_logits.view(
                    -1, rc_cnf.train.nclasses).to(self.device),
                labels.squeeze(1))

            batch_loss.backward()
            batch_loss = batch_loss / gradient_acc_steps

            total_loss += batch_loss.item() / current_batch_size

            (batch_acc, _, batch_precision, batch_f1,
                _, _, batch_stats_per_label) = self.evaluate_(
                classification_logits, labels, ignore_idx=-1)

            loss_per_batch.append(batch_loss / current_batch_size)
            accuracy_per_batch.append(batch_acc)

            torch.nn.utils.clip_grad_norm_(
                self.component.model.parameters(), max_grad_norm)

            if (i % gradient_acc_steps) == 0:
                self.component.optimizer.step()
                self.component.scheduler.step()
            if ((i + 1) % current_batch_size == 0):
                logger.debug(
                    "[Epoch: %d, loss per batch, accuracy per batch: %.3f,"
                    " %.3f, average total loss %.3f , total loss %.3f]",
                    epoch, loss_per_batch[-1], accuracy_per_batch[-1],
                    total_loss / (i + 1), total_loss)

            pbar.update(current_batch_size)

        pbar.close()

        losses_per_epoch = []
        accuracy_per_epoch = []
        f1_per_epoch = []
        if len(loss_per_batch) > 0:
            losses_per_epoch.append(
                sum(loss_per_batch) / len(loss_per_batch))
            logger.info("Losses at Epoch %d: %.5f" %
                        (epoch, losses_per_epoch[-1]))

        if len(accuracy_per_batch) > 0:
            accuracy_per_epoch.append(
                sum(accuracy_per_batch) / len(accuracy_per_batch))
            logger.info("Train accuracy at Epoch %d: %.5f" %
                        (epoch, accuracy_per_epoch[-1]))

        total_loss = total_loss / (i + 1)

        end_time = datetime.now().time()

        logger.info(
            "========================"
            " TRAIN SET TEST RESULTS "
            "========================")
        _ = self.evaluate_results(train_dataloader, self.component.pad_id)

        logger.info(
            "========================"
            " TEST SET TEST RESULTS "
            "========================")
        results = self.evaluate_results(
            test_dataloader, self.component.pad_id)

        f1_per_epoch.append(results['f1'])

        logger.info("Epoch finished, took %s seconds",
                    str(datetime.combine(date.today(), end_time)
                        - datetime.combine(date.today(), start_time)))

        self.epoch += 1

        if len(f1_per_epoch) > 0 and f1_per_epoch[-1] > self.best_f1:
            self.best_f1 = f1_per_epoch[-1]
            save_state(
                self.component.model, self.component.optimizer,
                self.component.scheduler, self.epoch, self.best_f1,
                checkpoint_path, model_name=rc_cnf.general.model_name,
                task=self.task, is_checkpoint=False)

        if (epoch % 1) == 0:
            save_results(
                {
                    "losses_per_epoch": losses_per_epoch,
                    "accuracy_per_epoch": accuracy_per_epoch,
                    "f1_per_epoch": f1_per_epoch,
                    "epoch": epoch
                }, file_prefix="train", path=checkpoint_path)
            save_state(self.component.model, self.component.optimizer,
                       self.component.scheduler, self.epoch, self.best_f1,
                       checkpoint_path,
                       model_name=rc_cnf.general.model_name,
                       task=self.task, is_checkpoint=True)
        return losses_per_epoch, accuracy_per_epoch, f1_per_epoch

    def evaluate_(self, output_logits, labels, ignore_idx):
        # ignore index (padding) when calculating accuracy
        idxs = (labels != ignore_idx).squeeze()
        labels_ = labels.squeeze()[idxs].to(self.device)
        pred_labels = torch.softmax(output_logits, dim=1).max(1)[1]
        pred_labels = pred_labels[idxs].to(self.device)

        true_labels = labels_.cpu().numpy().tolist(
        ) if labels_.is_cuda else labels_.numpy().tolist()
        pred_labels = pred_labels.cpu().numpy().tolist(
        ) if pred_labels.is_cuda else pred_labels.numpy().tolist()

        unique_labels = set(true_labels)

        batch_size = len(true_labels)

        stat_per_label = dict()

        total_tp, total_fp, total_tn, total_fn = 0, 0, 0, 0
        acc, micro_recall, micro_precision, micro_f1 = 0, 0, 0, 0

        for label in unique_labels:
            stat_per_label[label] = {
                "tp": 0, "fp": 0, "tn": 0, "fn": 0,
                "f1": 0.0, "acc": 0.0, "prec": 0.0, "recall": 0.0}

            for true_label_idx in range(len(true_labels)):
                if true_labels[true_label_idx] == label:
                    if pred_labels[true_label_idx] == label:
                        stat_per_label[label]["tp"] += 1
                        total_tp += 1
                    if pred_labels[true_label_idx] != label:
                        stat_per_label[label]["fp"] += 1
                        total_fp += 1
                elif (true_labels[true_label_idx] != label and
                      label == pred_labels[true_label_idx]):
                    stat_per_label[label]["fn"] += 1
                    total_fn += 1
                else:
                    stat_per_label[label]["tn"] += 1
                    total_tn += 1

            lbl_tp_tn = stat_per_label[label]["tn"] + \
                stat_per_label[label]["tp"]

            lbl_tp_fn = stat_per_label[label]["fn"] + \
                stat_per_label[label]["tp"]
            lbl_tp_fn = lbl_tp_fn if lbl_tp_fn > 0.0 else 1.0

            lbl_tp_fp = stat_per_label[label]["tp"] + \
                stat_per_label[label]["fp"]
            lbl_tp_fp = lbl_tp_fp if lbl_tp_fp > 0.0 else 1.0

            stat_per_label[label]["acc"] = lbl_tp_tn / batch_size
            stat_per_label[label]["prec"] = (stat_per_label[label]["tp"] /
                                             lbl_tp_fp)
            stat_per_label[label]["recall"] = (stat_per_label[label]["tp"] /
                                               lbl_tp_fn)

            lbl_re_pr = stat_per_label[label]["recall"] + \
                stat_per_label[label]["prec"]
            lbl_re_pr = lbl_re_pr if lbl_re_pr > 0.0 else 1.0

            stat_per_label[label]["f1"] = (
                2 * (stat_per_label[label]["recall"] *
                     stat_per_label[label]["prec"])) / lbl_re_pr

        tp_fn = total_fn + total_tp
        tp_fn = tp_fn if tp_fn > 0.0 else 1.0

        tp_fp = total_fp + total_tp
        tp_fp = tp_fp if tp_fp > 0.0 else 1.0

        micro_recall = total_tp / tp_fn
        micro_precision = total_tp / tp_fp

        re_pr = micro_recall + micro_precision
        re_pr = re_pr if re_pr > 0.0 else 1.0
        micro_f1 = (2 * (micro_recall * micro_precision)) / re_pr

        acc = total_tp / batch_size

        return (acc, micro_recall, micro_precision, micro_f1,
                pred_labels, true_labels, stat_per_label)

    def evaluate_results(self, data_loader, pad_id):
        logger.info("Evaluating test samples...")
        rc_cnf = self.component.relcat_config
        if (rc_cnf.train.class_weights is not None and
                rc_cnf.train.enable_class_weights):
            criterion = nn.CrossEntropyLoss(weight=torch.FloatTensor(
                rc_cnf.train.class_weights).to(self.device))
        else:
            criterion = nn.CrossEntropyLoss()

        total_loss, total_acc, total_f1, total_recall, total_precision = (
            0.0, 0.0, 0.0, 0.0, 0.0)
        all_batch_stats_per_label = []

        self.component.model.eval()

        for i, data in enumerate(data_loader):
            with torch.no_grad():
                token_ids, e1_e2_start, labels, _, _ = data
                attention_mask = (token_ids != pad_id).float().to(self.device)
                token_type_ids = torch.zeros(
                    (*token_ids.shape[:2],)).long().to(self.device)

                labels = labels.to(self.device)

                model_output, pred_classification_logits = (
                    self.component.model(token_ids,
                                         token_type_ids=token_type_ids,
                                         attention_mask=attention_mask,
                                         Q=None,
                                         e1_e2_start=e1_e2_start))

                batch_loss = criterion(pred_classification_logits.view(
                    -1, rc_cnf.train.nclasses).to(self.device),
                    labels.squeeze(1))
                total_loss += batch_loss.item()

                (batch_accuracy, batch_recall, batch_precision, batch_f1,
                 pred_labels, true_labels, batch_stats_per_label) = (
                    self.evaluate_(pred_classification_logits,
                                   labels, ignore_idx=-1))

                all_batch_stats_per_label.append(batch_stats_per_label)

                total_acc += batch_accuracy
                total_recall += batch_recall
                total_precision += batch_precision
                total_f1 += batch_f1

        final_stats_per_label = {}

        for batch_label_stats in all_batch_stats_per_label:
            for label_id, stat_dict in batch_label_stats.items():

                if label_id not in final_stats_per_label.keys():
                    final_stats_per_label[label_id] = stat_dict
                else:
                    for stat, score in stat_dict.items():
                        final_stats_per_label[label_id][stat] += score

        for label_id, stat_dict in final_stats_per_label.items():
            for stat_name, value in stat_dict.items():
                final_stats_per_label[label_id][stat_name] = value / (i + 1)

        total_loss = total_loss / (i + 1)
        total_acc = total_acc / (i + 1)
        total_precision = total_precision / (i + 1)
        total_f1 = total_f1 / (i + 1)
        total_recall = total_recall / (i + 1)

        results = {
            "loss": total_loss,
            "accuracy": total_acc,
            "precision": total_precision,
            "recall": total_recall,
            "f1": total_f1
        }

        logger.info("=" * 20 + " Evaluation Results " + "=" * 20)
        logger.info(" no. of batches:" + str(i + 1))
        for key in sorted(results.keys()):
            logger.info(" %s = %0.3f" % (key, results[key]))
        logger.info("-" * 23 + " class stats " + "-" * 23)
        for label_id, stat_dict in final_stats_per_label.items():
            logger.info(
                "label: %s | f1: %0.3f | prec : %0.3f | acc: %0.3f | "
                "recall: %0.3f ",
                rc_cnf.general.idx2labels[label_id],
                stat_dict["f1"],
                stat_dict["prec"],
                stat_dict["acc"],
                stat_dict["recall"]
            )
        logger.info("-" * 59)
        logger.info("=" * 59)

        return results

    def pipe(self, stream: Iterable[MutableDocument], *args, **kwargs
             ) -> Iterator[MutableDocument]:
        rc_cnf = self.component.relcat_config

        predict_rel_dataset = RelData(
            cdb=self.cdb, config=rc_cnf,
            tokenizer=self.component.tokenizer)

        self.component.model = self.component.model.to(self.device)

        for doc_id, doc in enumerate(stream, 0):
            predict_rel_dataset.dataset, _ = self._create_test_train_datasets(
                data=predict_rel_dataset.create_base_relations_from_doc(
                    doc, doc_id=str(doc_id)),
                split_sets=False)

            predict_dataloader = DataLoader(
                dataset=predict_rel_dataset, shuffle=False,
                batch_size=rc_cnf.train.batch_size,
                num_workers=0, collate_fn=self.component.padding_seq,
                pin_memory=rc_cnf.general.pin_memory)

            total_rel_found = len(
                predict_rel_dataset.dataset["output_relations"])
            rel_idx = -1

            logger.info("total relations for doc: " + str(total_rel_found))
            logger.info("processing...")

            pbar = tqdm(total=total_rel_found)

            for i, data in enumerate(predict_dataloader):
                with torch.no_grad():
                    token_ids, e1_e2_start, labels, _, _ = data

                    attention_mask = (
                        token_ids != self.component.pad_id
                        ).float().to(self.device)
                    token_type_ids = torch.zeros(
                        *token_ids.shape[:2]).long().to(self.device)

                    (model_output,
                     pred_classification_logits) = self.component.model(
                        token_ids, token_type_ids=token_type_ids,
                        attention_mask=attention_mask,
                        e1_e2_start=e1_e2_start)

                    for i, pred_rel_logits in enumerate(
                            pred_classification_logits):
                        rel_idx += 1

                        confidence = torch.softmax(
                            pred_rel_logits, dim=0).max(0)
                        predicted_label_id = int(confidence[1].item())

                        relations: list = doc.get_addon_data(  # type: ignore
                            "relations")
                        relations.append(
                            {
                                "relation": rc_cnf.general.idx2labels[
                                    predicted_label_id],
                                "label_id": predicted_label_id,
                                "ent1_text": predict_rel_dataset.dataset[
                                    "output_relations"][rel_idx][
                                    2],
                                "ent2_text": predict_rel_dataset.dataset[
                                    "output_relations"][rel_idx][
                                    3],
                                "confidence": float("{:.3f}".format(
                                    confidence[0])),
                                "start_ent_pos": "",
                                "end_ent_pos": "",
                                "start_entity_id":
                                    predict_rel_dataset.dataset[
                                        "output_relations"][rel_idx][8],
                                "end_entity_id":
                                    predict_rel_dataset.dataset[
                                        "output_relations"][rel_idx][9]
                            })
                    pbar.update(len(token_ids))
            pbar.close()

            yield doc

    def predict_text_with_anns(self, text: str, annotations: list[dict]
                               ) -> MutableDocument:
        """ Creates spacy doc from text and annotation input.
        Predicts using self.__call__

        Args:
            text (str): text
            annotations (dict): dict containing the entities from NER
                (of your choosing), the format must be the following format:
                    [
                        {
                            "cui": "202099003", -this is optional
                            "value": "discoid lateral meniscus",
                            "start": 294,
                            "end": 318
                        },
                        {
                            "cui": "202099003",
                            "value": "Discoid lateral meniscus",
                            "start": 1905,
                            "end": 1929,
                        }
                    ]

        Returns:
            Doc: spacy doc with the relations.
        """
        # NOTE: This runs not an empty language, but the specified one
        base_tokenizer = create_tokenizer(
            self.cdb.config.general.nlp.provider, self.cdb.config)
        doc = base_tokenizer(text)

        for ann in annotations:
            tkn_idx = []
            for ind, word in enumerate(doc):
                end_char = word.base.char_index + len(word.base.text)
                if end_char <= ann['end'] and end_char > ann['start']:
                    tkn_idx.append(ind)
            entity = base_tokenizer.create_entity(
                doc, min(tkn_idx), max(tkn_idx) + 1, label=ann["value"])
            entity.cui = ann["cui"]
            doc.ner_ents.append(entity)

        doc = self(doc)

        return doc



from typing import Any, Callable

import xxhash
import dill
from io import BytesIO as StringIO


def dumps(obj: Any, length: bool = False) -> bytes:
    """Dump the content of an object to bytes.

    This method uses dill to dump the contents of an
    object into a BytesIO object and then either
    reads its bytes or (or length == True) simply
    reruns the process on the length of the byte array.

    Args:
        obj (Any): The object to dump.
        length (bool, optional):
            Whether to only dump the length of the file array.
            Defaults to False.

    Returns:
        bytes: The resulting byte array.
    """
    with StringIO() as file:
        dill.dump(obj, file, recurse=True)
        if length:
            return dumps(len(file.getvalue()), length=False)
        else:
            return file.getvalue()


class Hasher:
    """A consistent hasher.

    This class is able to hash the same object(s) to the
    same value every time.
    This is in contrast to the normal hashing in python
    that does not guarantee identical results over multiple
    runs.

    Args:
        dumper (Callable[[Any, bool], bytes], optional):
            The dumper to be used. Defaults to the `dumps` method.
    """

    def __init__(self, dumper: Callable[[Any, bool], bytes] = dumps):
        self.m = xxhash.xxh64()
        self._dumper = dumper

    def update(self, obj: Any, length: bool = False) -> None:
        """Update the hasher with the object in question.

        If `length = True` is passed, only the length of
        the byte array corresponding to the data is considered
        Otherwise the entire byte array is used.

        Args:
            obj (Any): The object to be added / hashed.
            length (bool, optional):
                Whether to only dump the length of the file array.
                Defaults to False.
        """
        self.m.update(self._dumper(obj, length))

    def update_bytes(self, b: bytes) -> None:
        """Update the hasher with a byte array.

        Args:
            b (bytes): The byte array to update with.
        """
        self.m.update(b)

    def hexdigest(self) -> str:
        """Get the hex for the current hash state.

        Returns:
            str: The hex representation of the hashed objects.
        """
        return self.m.hexdigest()



import pandas as pd
import numpy as np
import datetime
import logging
import re
from typing import Optional, Union, Any

from medcat.pipeline.pipeline import Pipeline
from medcat.cdb import CDB
from medcat.config import Config
from medcat.preprocessors.cleaners import prepare_name

PH_REMOVE = re.compile(r"(\s)\([a-zA-Z]+[^\)\(]*\)($)")


logger = logging.getLogger(__name__)


class CDBMaker(object):
    """Given a CSV it creates a CDB or updates an existing one.

    The CSV is expected to have

    Args:
        config (medcat.config.Config):
            Global config for MedCAT.
        cdb (medcat.cdb.CDB):
            If set the `CDBMaker` will update the existing `CDB` with
            new concepts in the CSV (Default value `None`).
    """

    def __init__(self, config: Config, cdb: Optional[CDB] = None) -> None:
        self.config = config
        # Set log level
        logger.setLevel(self.config.general.log_level)

        # To make life a bit easier
        self.cnf_cm = config.cdb_maker

        if cdb is None:
            self.cdb = CDB(config=self.config)
        else:
            self.cdb = cdb

        # Build the required spacy pipeline
        self.pipeline = Pipeline(self.cdb, vocab=None, model_load_path=None)

    def reset_cdb(self) -> None:
        """This will re-create a new internal CDB based on the same config.

        This will be necessary if/when you're wishing to call `prepare_csvs`
        multiple times on the same object `CDBMaker` instance.
        """
        self.cdb = CDB(config=self.config)

    def prepare_csvs(self,
                     csv_paths: Union[pd.DataFrame, list[str]],
                     sep: str = ',',
                     encoding: Optional[str] = None,
                     escapechar: Optional[str] = None,
                     index_col: bool = False,
                     full_build: bool = False,
                     only_existing_cuis: bool = False, **kwargs: Any) -> CDB:
        r"""Compile one or multiple CSVs into a CDB.

        Note: This class/method generally uses the same instance of the CDB.
              So if you're using the same CDBMaker and calling `prepare_csvs`
              multiple times, you are likely to get leakage from prior calls
              into new ones.
              To reset the CDB, call `reset_cdb`.

        Args:
            csv_paths (Union[pd.DataFrame, list[str]]):
                An array of paths to the csv files that should be processed.
                Can also be an array of pd.DataFrames
            sep (str):
                If necessary a custom separator for the csv files
                Defaults to ','.
            encoding (Optional[str]):
                Encoding to be used for reading the CSV file
                Defaults to `None`.
            escapechar (Optional[str]):
                Escape char for the CSV (Default value None).
            index_col (bool):
                Index column for pandas read_csv (Default value False).
            full_build (bool):
                If False only the core portions of the CDB will be built (the
                ones required for the functioning of MedCAT). If True,
                everything will be added to the CDB - this
                usually includes concept descriptions, various forms of names
                etc (take care that this option produces a much larger CDB).
                Defaults to False.
            only_existing_cuis (bool):
                If True no new CUIs will be added, but only linked names will
                be extended. Mainly used when enriching names of a CDB (e.g.
                SNOMED with UMLS terms). Default to `False`.
            kwargs (Any):
                Will be passed to pandas for CSV reading

        Note:
            \*\*kwargs:
                Will be passed to pandas for CSV reading

        Returns:
            CDB: CDB with the new concepts added.
        """

        useful_columns = ['cui', 'name', 'ontologies', 'name_status',
                          'type_ids', 'description']
        name_status_options = {'A', 'P', 'N'}
        pn_cnf_parts = (self.config.general, self.config.preprocessing,
                        self.config.cdb_maker)

        for csv_path in csv_paths:
            # Read CSV, everything is converted to strings
            if isinstance(csv_path, str):
                logger.info("Started importing concepts from: {}".format(
                    csv_path))
                df = pd.read_csv(csv_path, sep=sep, encoding=encoding,
                                 escapechar=escapechar, index_col=index_col,
                                 dtype=str, **kwargs)
            else:
                # Not very clear, but csv_path can be a pre-loaded csv
                df = csv_path
            df = df.fillna('')

            # Find which columns to use from the CSV
            cols: list = []
            col2ind = {}
            for col in list(df.columns):
                if str(col).lower().strip() in useful_columns:
                    col2ind[str(col).lower().strip()] = len(cols)
                    cols.append(col)

            _time = None  # Used to check speed
            _logging_freq = np.ceil(len(df[cols]) / 100)
            multi_sep = self.cnf_cm.multi_separator
            for row_id, row in enumerate(df[cols].values):
                if row_id % _logging_freq == 0:
                    # Print some stats
                    if _time is None:
                        # Add last time if it does not exist
                        _time = datetime.datetime.now()
                    # Get current time
                    ctime = datetime.datetime.now()
                    # Get time difference
                    timediff = ctime - _time
                    logger.info(
                        "Current progress: %.0f%% at %.3fs per %d rows",
                        (row_id / len(df)) * 100,
                        timediff.microseconds / 10**6 + timediff.seconds,
                        (len(df[cols]) // 100))
                    # Set previous time to current time
                    _time = ctime

                # This must exist
                cui = row[col2ind['cui']].strip().upper()

                if not only_existing_cuis or (only_existing_cuis and
                                              cui in self.cdb.cui2info):
                    if 'ontologies' in col2ind:
                        ontologies = set(
                            [ontology.strip()
                             for ontology in row[col2ind['ontologies']
                                                 ].upper().split(multi_sep)
                             if len(ontology.strip()) > 0])
                    else:
                        ontologies = set()

                    if 'name_status' in col2ind:
                        name_status = row[col2ind['name_status']
                                          ].strip().upper()

                        # Must be allowed
                        if name_status not in name_status_options:
                            name_status = 'A'
                    else:
                        # Defaults to A - meaning automatic
                        name_status = 'A'

                    if 'type_ids' in col2ind:
                        type_ids = set(
                            [type_id.strip()
                             for type_id in row[col2ind['type_ids']
                                                ].upper().split(multi_sep)
                             if len(type_id.strip()) > 0])
                    else:
                        type_ids = set()

                    # Get the ones that do not need any changing
                    if 'description' in col2ind:
                        description = row[col2ind['description']].strip()
                    else:
                        description = ""

                    # We can have multiple versions of a name
                    # {'name': {'tokens': [<str>], 'snames': [<str>]}}
                    names: dict = {}

                    raw_names = [
                        raw_name.strip()
                        for raw_name in row[col2ind['name']].split(multi_sep)
                        if len(raw_name.strip()) > 0]
                    for raw_name in raw_names:
                        raw_name = raw_name.strip()
                        prepare_name(
                            raw_name, self.pipeline.tokenizer_with_tag,
                            names, pn_cnf_parts)

                        if (self.config.cdb_maker.remove_parenthesis > 0 and
                                name_status == 'P'):
                            # Should we remove the content in parenthesis
                            # from primary names and add them also
                            raw_name = PH_REMOVE.sub(" ", raw_name).strip()
                            if len(raw_name) >= self.cnf_cm.remove_parenthesis:
                                prepare_name(
                                    raw_name, self.pipeline.tokenizer_with_tag,
                                    names, pn_cnf_parts)

                    self.cdb._add_concept(
                        cui=cui, names=names, ontologies=ontologies,
                        name_status=name_status, type_ids=type_ids,
                        description=description, full_build=full_build)
                    # DEBUG
                    logger.debug(
                        "\n\n**** Added\n CUI: %s\n Names: %s\n Ontologies: %s"
                        "\n Name status: %s\n Type IDs: %s\n Description: %s\n"
                        " Is full build: %s",
                        cui, names, ontologies, name_status, type_ids,
                        description, full_build)

        return self.cdb



import os
import json
import logging
import datasets
import torch
from datetime import datetime
from typing import Iterable, Iterator, Optional, Union, Callable
from typing import cast
import inspect
from functools import partial

from medcat.cdb.cdb import CDB
from medcat.components.addons.meta_cat.ml_utils import set_all_seeds
from medcat.utils.ner import transformers_ner
from medcat.utils.postprocessing import create_main_ann
from medcat.utils.hasher import Hasher
from medcat.config.config_transformers_ner import ConfigTransformersNER
from medcat.config.config import ComponentConfig
from medcat.components.ner.trf.tokenizer import (
    TransformersTokenizer)
from medcat.utils.ner.metrics import metrics
from medcat.utils.ner.data_collator import CollateAndPadNER
from medcat.tokenizing.tokens import MutableDocument, MutableEntity
from medcat.tokenizing.tokenizers import BaseTokenizer
from medcat.storage.serialisers import (
    serialise, AvailableSerialisers, deserialise)
from medcat.storage.serialisables import SerialisingStrategy
from medcat.preprocessors.cleaners import NameDescriptor
from medcat.components.types import CoreComponentType, AbstractCoreComponent
from medcat.vocab import Vocab
from medcat.utils.defaults import COMPONENTS_FOLDER

from transformers import (
    Trainer, AutoModelForTokenClassification, AutoTokenizer)
from transformers import pipeline, TrainingArguments
from transformers.trainer_callback import TrainerCallback

# It should be safe to do this always, as all other multiprocessing
# will be finished before data comes to meta_cat
os.environ["TOKENIZERS_PARALLELISM"] = "true"
os.environ['WANDB_DISABLED'] = 'true'


logger = logging.getLogger(__name__)


class TransformersNER(AbstractCoreComponent):
    name = 'transformers_ner'
    _def_serialiser = AvailableSerialisers.dill

    def __init__(self, cdb: CDB,
                 base_tokenizer: BaseTokenizer,
                 component: 'TransformersNERComponent',
                 config: Optional[ConfigTransformersNER] = None,
                 training_arguments=None,) -> None:
        self._component = component

    @classmethod
    def create_new(cls, cdb: CDB, base_tokenizer: BaseTokenizer,
                   config: Optional[ConfigTransformersNER] = None,
                   training_arguments=None) -> 'TransformersNER':
        comp = TransformersNERComponent(
                cdb, base_tokenizer, config, training_arguments)
        return cls(cdb=cdb, base_tokenizer=base_tokenizer,
                   config=config, training_arguments=training_arguments,
                   component=comp)

    @classmethod
    def create_new_component(
            cls, cnf: ComponentConfig, tokenizer: BaseTokenizer,
            cdb: CDB, vocab: Vocab, model_load_path: Optional[str]
            ) -> 'TransformersNER':
        config = cdb.config.components.ner.custom_cnf
        if not isinstance(config, ConfigTransformersNER):
            raise ValueError(
                "Did not find correct Transformers NER config. "
                f"Found: {config}")
        # TODO: anywhere to get these?
        training_arguments = None
        if model_load_path is not None:
            load_path = os.path.join(
                model_load_path, COMPONENTS_FOLDER, cls.NAME_PREFIX + "ner")
            return cls.load_existing(cdb, tokenizer, load_path,
                                     training_arguments, config)
        return cls.create_new(cdb, tokenizer, config, training_arguments)

    @classmethod
    def load_existing(cls, cdb: CDB, base_tokenizer: BaseTokenizer,
                      load_path: str, training_arguments=None,
                      config: Optional[ConfigTransformersNER] = None,
                      ) -> 'TransformersNER':
        comp = _load_component(cdb, load_path, base_tokenizer)
        return cls(cdb=cdb, base_tokenizer=base_tokenizer,
                   config=config, training_arguments=training_arguments,
                   component=comp)

    def get_type(self):
        return CoreComponentType.ner

    @property
    def should_save(self) -> bool:
        return True

    def save(self, folder: str, overwrite: bool = False) -> None:
        _save_component(self._component,
                        folder, serialiser=self._def_serialiser,
                        overwrite=overwrite)

    def __call__(self, doc: MutableDocument) -> MutableDocument:
        return self._component(doc)

    # for manual serialisability

    def get_folder_name(self) -> str:
        return self.NAME_PREFIX + self.get_type().name

    def serialise_to(self, folder_path: str) -> None:
        self.save(folder_path)

    @classmethod
    def deserialise_from(cls, folder_path: str, **init_kwargs
                         ) -> 'TransformersNER':
        return cls.load_existing(
            load_path=folder_path,
            cdb=init_kwargs['cdb'],
            base_tokenizer=init_kwargs['tokenizer'],
            # from Config.components.ner (of type Ner)
            config=init_kwargs['cnf'].custom_cnf)

    def get_strategy(self) -> SerialisingStrategy:
        return SerialisingStrategy.MANUAL

    @classmethod
    def get_init_attrs(cls) -> list[str]:
        return []

    @classmethod
    def ignore_attrs(cls) -> list[str]:
        return []

    @classmethod
    def include_properties(cls) -> list[str]:
        return []


def _save_component(
        comp: 'TransformersNERComponent', save_dir_path: str,
        serialiser: AvailableSerialisers = AvailableSerialisers.dill,
        overwrite: bool = False,
        ) -> None:
    """Save all components of this class to a file

    Args:
        save_dir_path (str):
            Path to the directory where everything will be saved.
        serialiser (AvailableSerialisers):
            The serialiser type to use.
    """
    # Create dirs if they do not exist
    os.makedirs(save_dir_path, exist_ok=True)

    # Save tokenizer
    comp.tokenizer.save(os.path.join(save_dir_path, 'tokenizer.dat'))

    # Save config
    folder = os.path.join(save_dir_path, 'cat_config')
    if not os.path.exists(folder):
        os.mkdir(folder)
    serialise(serialiser, comp.config, folder, overwrite=overwrite)

    # Save the model
    comp.model.save_pretrained(save_dir_path)

    # Save the cdb
    folder = os.path.join(save_dir_path, 'CDB')
    if not os.path.exists(folder):
        os.mkdir(folder)
    serialise(serialiser, comp.cdb, folder, overwrite=overwrite)


def _load_component(cdb: CDB, save_dir_path: str,
                    base_tokenizer: BaseTokenizer,
                    config_dict: Optional[dict] = None
                    ) -> 'TransformersNERComponent':
    """Load a meta_cat object.

    Args:
        save_dir_path (str):
            The directory where all was saved.
        base_tokenizer (BaseTokenizer):
            The tokenizer for the model pack (for implementation).
        config_dict (dict):
            This can be used to overwrite saved parameters for
            this TransformersNER instance. Why? It is needed in
            certain cases where we autodeploy stuff.

    Returns:
        TransformersNER:
            The TNER instance.
    """

    # Load config
    config = cast(ConfigTransformersNER, deserialise(
        os.path.join(save_dir_path, 'cat_config')))
    config.general.model_name = save_dir_path

    # Overwrite loaded parameters with something new
    if config_dict is not None:
        config.merge_config(config_dict)

    # Load cdb
    # cdb = cast(CDB, deserialise(os.path.join(save_dir_path, 'cdb')))

    config.general.model_name = save_dir_path
    ner = TransformersNERComponent(cdb=cdb, config=config,
                                   base_tokenizer=base_tokenizer)
    ner.create_eval_pipeline()
    return ner


TrCBCreator = Callable[[Trainer], TrainerCallback]


class TransformersNERComponent:
    """TODO: Add documentation"""

    def __init__(self, cdb: CDB,
                 base_tokenizer: BaseTokenizer,
                 config: Optional[ConfigTransformersNER] = None,
                 training_arguments=None) -> None:
        self.base_tokenizer = base_tokenizer
        self.cdb = cdb
        if config is None:
            cnf = cdb.config.components.ner.custom_cnf
            if cnf is not None and isinstance(cnf, ConfigTransformersNER):
                config = cnf
            else:
                config = ConfigTransformersNER()

        self.config = config
        set_all_seeds(config.general.seed)

        self.model = AutoModelForTokenClassification.from_pretrained(
            config.general.model_name)

        # Get the tokenizer either create a new one or load existing
        if os.path.exists(os.path.join(
                config.general.model_name, 'tokenizer.dat')):
            self.tokenizer = TransformersTokenizer.load(
                os.path.join(config.general.model_name, 'tokenizer.dat'))
        else:
            hf_tokenizer = AutoTokenizer.from_pretrained(
                self.config.general.model_name)
            self.tokenizer = TransformersTokenizer(hf_tokenizer)

        if training_arguments is None:
            self.training_arguments = TrainingArguments(
                output_dir='./results',
                # directory for storing logs
                logging_dir='./logs',
                # total number of training epochs
                num_train_epochs=10,
                # batch size per device during training
                per_device_train_batch_size=1,
                # batch size for evaluation
                per_device_eval_batch_size=1,
                # strength of weight decay
                weight_decay=0.14,
                warmup_ratio=0.01,
                # Should be smaller when finetuning an existing deid model
                learning_rate=4.47e-05,
                eval_accumulation_steps=1,
                # We want to get to bs=4
                gradient_accumulation_steps=4,
                do_eval=True,
                # eval_strategy since transformers==4.41
                eval_strategy='epoch',
                logging_strategy='epoch',     # type: ignore
                save_strategy='epoch',        # type: ignore
                # Can be changed if our preference is not recall but precision
                # or f1
                metric_for_best_model='eval_recall',
                load_best_model_at_end=True,
                remove_unused_columns=False)
        else:
            self.training_arguments = training_arguments

    def create_eval_pipeline(self):

        if self.config.general.chunking_overlap_window is None:
            logger.warning(
                "Chunking overlap window attribute in the config is set to "
                "None, hence chunking is disabled. Be cautious, PII data MAY "
                "BE REVEALED. To enable chunking, set the value to 0 or above")
        self.ner_pipe = pipeline(
            model=self.model, task="ner",
            tokenizer=self.tokenizer.hf_tokenizer,
            stride=self.config.general.chunking_overlap_window)
        if not hasattr(self.ner_pipe.tokenizer, '_in_target_context_manager'):
            # NOTE: this will fix the DeID model(s) created before medcat 1.9.3
            #       though this fix may very well be unstable
            self.ner_pipe.tokenizer._in_target_context_manager = False
        if not hasattr(self.ner_pipe.tokenizer, 'split_special_tokens'):
            # NOTE: this will fix the DeID model(s) created with transformers
            #       before 4.42 and allow them to run with later transformers
            self.ner_pipe.tokenizer.split_special_tokens = False
        if (not hasattr(self.ner_pipe.tokenizer, 'pad_token') and
                hasattr(self.ner_pipe.tokenizer, '_pad_token')):
            # NOTE: This will fix the DeID model(s) created with transformers
            #       before 4.47 and allow them to run with later transformmers
            #       versions
            #       In 4.47 the special tokens started to be used differently,
            #       yet our saved model is not aware of that. So we need to
            #       explicitly fix that.
            special_tokens_map = self.ner_pipe.tokenizer.__dict__.get(
                '_special_tokens_map', {})
            for name in self.ner_pipe.tokenizer.SPECIAL_TOKENS_ATTRIBUTES:
                # previously saved in (e.g) _pad_token
                prev_val = getattr(self.ner_pipe.tokenizer, f"_{name}")
                # now saved in the special tokens map by its name
                special_tokens_map[name] = prev_val
            # the map is saved in __dict__ explicitly, and
            # it is later used in __getattr__ of the base class.
            self.ner_pipe.tokenizer.__dict__[
                '_special_tokens_map'] = special_tokens_map

        self.ner_pipe.device = self.model.device

    def get_hash(self) -> str:
        """A partial hash trying to catch differences between models.

        Returns:
            str: The hex hash.
        """
        hasher = Hasher()
        # Set last_train_on if None
        if self.config.general.last_train_on is None:
            self.config.general.last_train_on = datetime.now().timestamp()

        hasher.update(self.config.get_hash())
        return hasher.hexdigest()

    def _prepare_dataset(self, json_path, ignore_extra_labels,
                         meta_requirements, file_name='data.json'):
        def merge_data_loaded(base, other):
            if not base:
                return other
            elif other is None:
                return base
            else:
                for p in other['projects']:
                    base['projects'].append(p)
            return base

        if isinstance(json_path, str):
            json_path = [json_path]

        # Merge data from all different data paths
        data_loaded: dict = {}
        for path in json_path:
            with open(path, 'r') as f:
                data_loaded = merge_data_loaded(data_loaded, json.load(f))

        # Remove labels that did not exist in old dataset
        if ignore_extra_labels and self.tokenizer.label_map:
            logger.info("Ignoring extra labels from the data")
            for p in data_loaded['projects']:
                for d in p['documents']:
                    new_anns = []
                    for a in d['annotations']:
                        if a['cui'] in self.tokenizer.label_map:
                            new_anns.append(a)
                    d['annotations'] = new_anns
        if meta_requirements is not None:
            logger.info("Removing anns that do not meet meta requirements")
            for p in data_loaded['projects']:
                for d in p['documents']:
                    new_anns = []
                    for a in d['annotations']:
                        if all([a['meta_anns'][name]['value'] == value
                                for name, value in meta_requirements.items()]):
                            new_anns.append(a)
                    d['annotations'] = new_anns

        # Here we have to save the data because of the data loader
        os.makedirs('results', exist_ok=True)
        out_path = os.path.join(os.getcwd(), 'results', file_name)
        json.dump(data_loaded, open(out_path, 'w'))

        return out_path

    def train(self,
              json_path: Union[str, list, None] = None,
              ignore_extra_labels=False,
              dataset=None,
              meta_requirements=None,
              train_json_path: Union[str, list, None] = None,
              test_json_path: Union[str, list, None] = None,
              trainer_callbacks: Optional[list[TrCBCreator]] = None
              ) -> tuple:
        """Train or continue training a model give a json_path containing a
        MedCATtrainer export. It will continue training if an existing model
        is loaded or start new training if the model is blank/new.

        Args:
            json_path (str or list):
                Path/Paths to a MedCATtrainer export containing the
                meta_annotations we want to train for.
            ignore_extra_labels:
                Makes only sense when an existing deid model was loaded and
                from the new data we want to ignore labels that did not exist
                in the old model.
            dataset: Defaults to None.
            meta_requirements: Defaults to None
            train_json_path (Union[str, list, None]):
                The json path for the training data. Defaults to None.
            test_json_path (Union[str, list, None]):
                The json path for the test data. Defaults to None.
            trainer_callbacks (list[TrCBCreator]):
                A list of trainer callbacks for collecting metrics during the
                training at the client side. The transformers Trainer object
                will be passed in when each callback is called.

        Returns:
            Tuple: The dataframe, examples, and the dataset
        """

        if dataset is None:
            # Load the medcattrainer export
            if json_path is not None:
                json_path = self._prepare_dataset(
                    json_path, ignore_extra_labels=ignore_extra_labels,
                    meta_requirements=meta_requirements,
                    file_name='data_eval.json')
            elif test_json_path is not None and train_json_path is not None:
                train_json_path = self._prepare_dataset(
                    train_json_path, ignore_extra_labels=ignore_extra_labels,
                    meta_requirements=meta_requirements,
                    file_name='data_train.json')
                test_json_path = self._prepare_dataset(
                    test_json_path, ignore_extra_labels=ignore_extra_labels,
                    meta_requirements=meta_requirements,
                    file_name='data_test.json')
            # Load dataset

            # NOTE: The following is for backwards comppatibility
            #       in datasets==2.20.0 `trust_remote_code=True`
            #       must be explicitly specified, otherwise an error is raised.
            #       On the other hand, the keyword argument was added in
            #       datasets==2.16.0 yet we support datasets>=2.2.0.
            #       So we need to use the kwarg if applicable and omit
            #       its use otherwise.
            if func_has_kwarg(datasets.load_dataset, 'trust_remote_code'):
                ds_load_dataset = partial(datasets.load_dataset,
                                          trust_remote_code=True)
            else:
                ds_load_dataset = datasets.load_dataset
            if json_path:
                dataset = ds_load_dataset(os.path.abspath(
                    transformers_ner.__file__),
                    data_files={'train': json_path},  # type: ignore
                    split='train',
                    cache_dir='/tmp/')
                # We split before encoding so the split is document level,
                # as encoding  does the document splitting into max_seq_len
                dataset = dataset.train_test_split(
                    test_size=self.config.general.test_size)  # type: ignore
            elif train_json_path and test_json_path:
                dataset = ds_load_dataset(
                    os.path.abspath(transformers_ner.__file__),
                    data_files={
                        'train': train_json_path,
                        'test': test_json_path},  # type: ignore
                    cache_dir='/tmp/')
            else:
                raise ValueError(
                    "Either json_path or train_json_path and test_json_path "
                    "must be provided when no dataset is provided")

        # Update labelmap in case the current dataset has more labels
        # than what we had before
        self.tokenizer.calculate_label_map(dataset['train'])
        self.tokenizer.calculate_label_map(dataset['test'])

        if self.model.num_labels != len(self.tokenizer.label_map):
            logger.warning(
                "The dataset contains labels we've not seen before, "
                "model is being reinitialized")
            logger.warning("Model: %s vs Dataset: %s",
                           self.model.num_labels,
                           len(self.tokenizer.label_map))
            self.model = AutoModelForTokenClassification.from_pretrained(
                self.config.general.model_name,
                num_labels=len(self.tokenizer.label_map),
                ignore_mismatched_sizes=True)
            self.tokenizer.cui2name = {
                k: self.cdb.get_name(k)
                for k in self.tokenizer.label_map.keys()}

        self.model.config.id2label = {
            v: k for k, v in self.tokenizer.label_map.items()}
        self.model.config.label2id = self.tokenizer.label_map

        # Encode dataset
        # Note: tokenizer.encode performs chunking
        encoded_dataset = dataset.map(
                lambda examples: self.tokenizer.encode(
                    examples, ignore_subwords=False),
                batched=True,
                remove_columns=['ent_cuis', 'ent_ends', 'ent_starts', 'text'])

        data_collator = CollateAndPadNER(
            self.tokenizer.hf_tokenizer.pad_token_id)  # type: ignore
        trainer = Trainer(
                model=self.model,
                args=self.training_arguments,
                train_dataset=encoded_dataset['train'],
                eval_dataset=encoded_dataset['test'],
                compute_metrics=lambda p: metrics(
                    p, tokenizer=self.tokenizer,
                    dataset=encoded_dataset['test'],
                    verbose=self.config.general.verbose_metrics),
                data_collator=data_collator,  # type: ignore
                tokenizer=None)
        if trainer_callbacks:
            for tr_callback in trainer_callbacks:
                tcbo = tr_callback(trainer)
                # NOTE: No idea why mypy isn't able to find the method
                #       It reports (`[attr-defined]`):
                #          error: "Trainer" has no attribute "callback_handler"
                trainer.add_callback(tcbo)  # type: ignore

        # NOTE: No idea why mypy isn't able to find the method
        #       It reports:
        #          error: "Trainer" has no attribute "train"  [attr-defined]
        trainer.train()  # type: ignore

        # Save the training time
        self.config.general.last_train_on = datetime.now().timestamp()

        output_dir = self.training_arguments.output_dir
        if output_dir is None:
            # NOTE: shouldn't ever really happen
            raise ValueError("Unable to save output during training "
                             "since output path is None")
        # Save everything
        _save_component(self, save_dir_path=os.path.join(
            output_dir, 'final_model'),
            overwrite=True)

        # Run an eval step and return metrics
        p = trainer.predict(encoded_dataset['test'])  # type: ignore
        df, examples = metrics(p, return_df=True, tokenizer=self.tokenizer,
                               dataset=encoded_dataset['test'])

        # Create the pipeline for eval
        self.create_eval_pipeline()

        return df, examples, dataset

    def eval(self, json_path: Union[str, list, None] = None, dataset=None,
             ignore_extra_labels=False, meta_requirements=None):
        if dataset is None:
            json_path = self._prepare_dataset(
                json_path, ignore_extra_labels=ignore_extra_labels,
                meta_requirements=meta_requirements,
                file_name='data_eval.json')
            # Load dataset
            dataset = datasets.load_dataset(
                os.path.abspath(transformers_ner.__file__),
                data_files={'train': json_path},  # type: ignore
                split='train',
                cache_dir='/tmp/')

        # Encode dataset
        # Note: tokenizer.encode performs chunking
        encoded_dataset = dataset.map(
                lambda examples: self.tokenizer.encode(
                    examples, ignore_subwords=False),
                batched=True,
                remove_columns=['ent_cuis', 'ent_ends', 'ent_starts', 'text'])

        data_collator = CollateAndPadNER(
            self.tokenizer.hf_tokenizer.pad_token_id)  # type: ignore
        # TODO: switch from trainer to model prediction
        trainer = Trainer(
                model=self.model,
                args=self.training_arguments,
                train_dataset=None,
                eval_dataset=encoded_dataset,  # type: ignore
                compute_metrics=None,
                data_collator=data_collator,  # type: ignore
                tokenizer=None)

        # Run an eval step and return metrics
        p = trainer.predict(encoded_dataset)  # type: ignore
        df, examples = metrics(p, return_df=True, tokenizer=self.tokenizer,
                               dataset=encoded_dataset)

        return df, examples, dataset

    def expand_model_with_concepts(self, cui2preferred_name: dict[str, str],
                                   use_avg_init: bool = True) -> None:
        """Expand the model with new concepts and their preferred names, which
        requires subsequent retraining on the model.

        Args:
            cui2preferred_name(Dict[str, str]):
                Dictionary where each key is the literal ID of the concept to
                be added and each value is its preferred name.
            use_avg_init(bool):
                Whether to use the average of existing weights or biases as
                the initial value for the new concept. Defaults to True.
        """

        avg_weight = torch.mean(self.model.classifier.weight, dim=0,
                                keepdim=True)
        avg_bias = torch.mean(self.model.classifier.bias, dim=0, keepdim=True)

        new_cuis = set()
        for label, preferred_name in cui2preferred_name.items():
            if label in self.model.config.label2id.keys():
                logger.warning(
                    "Concept ID '%s' already exists in the model, skipping...",
                    label)
                continue

            sname = preferred_name.lower().replace(" ", "~")
            new_names = {
                sname: NameDescriptor(
                    tokens=[],
                    snames={sname},
                    raw_name=preferred_name,
                    is_upper=True
                )
            }
            self.cdb.add_names(
                cui=label, names=new_names, name_status="P", full_build=True)

            new_label_id = sorted(self.model.config.label2id.values())[-1] + 1
            self.model.config.label2id[label] = new_label_id
            self.model.config.id2label[new_label_id] = label
            self.tokenizer.label_map[label] = new_label_id
            self.tokenizer.cui2name = {k: self.cdb.get_name(k) for
                                       k in self.tokenizer.label_map.keys()}

            if use_avg_init:
                self.model.classifier.weight = torch.nn.Parameter(
                    torch.cat((self.model.classifier.weight, avg_weight), 0)
                )
                self.model.classifier.bias = torch.nn.Parameter(
                    torch.cat((self.model.classifier.bias, avg_bias), 0)
                )
            else:
                self.model.classifier.weight = torch.nn.Parameter(
                    torch.cat((self.model.classifier.weight, torch.randn(
                        1, self.model.config.hidden_size)), 0)
                )
                self.model.classifier.bias = torch.nn.Parameter(
                    torch.cat((self.model.classifier.bias, torch.randn(1)), 0)
                )
            self.model.num_labels += 1
            self.model.classifier.out_features += 1

            new_cuis.add(label)

        logger.info("Model expanded with the new concept(s): %s and shall be "
                    "retrained before use.", str(new_cuis))

    @staticmethod
    def batch_generator(stream: Iterable[MutableDocument],
                        batch_size_chars: int
                        ) -> Iterable[list[MutableDocument]]:
        docs = []
        char_count = 0
        for doc in stream:
            char_count += len(doc.base.text)
            docs.append(doc)
            if char_count < batch_size_chars:
                continue
            yield docs
            docs = []
            char_count = 0

        # If there is anything left return that also
        if len(docs) > 0:
            yield docs

    def pipe(self, stream: Iterable[Union[MutableDocument, None]],
             *args, **kwargs) -> Iterator[MutableDocument]:
        """Process many documents at once.

        Args:
            stream (Iterable[MutableDocument]):
                List of documents.
            *args: Extra arguments (not used here).
            **kwargs: Extra keyword arguments (not used here).

        Yields:
            Doc: The same document.

        Returns:
            Iterator[MutableDocument]: If the stream is None or empty.
        """
        # Just in case
        if stream is None or not stream:
            # return an empty generator
            return

        batch_size_chars = self.config.general.pipe_batch_size_in_chars
        yield from self._process(stream, batch_size_chars)  # type: ignore

    def _process_doc(self, doc: MutableDocument):
        aggr_strat = self.config.general.ner_aggregation_strategy
        res = self.ner_pipe(doc.base.text,
                            aggregation_strategy=aggr_strat)
        doc.ner_ents = []  # type: ignore
        for r in res:
            inds = []
            for ind, word in enumerate(doc):
                end_char = word.base.char_index + len(word.base.text)
                if end_char <= r['end'] and end_char > r['start']:
                    inds.append(ind)
                # To not loop through everything
                if end_char > r['end']:
                    break
            if not inds:
                continue

            entity: MutableEntity = self.base_tokenizer.create_entity(
                doc, min(inds), max(inds) + 1,
                label=r['entity_group'])
            entity.cui = r['entity_group']
            entity.context_similarity = r['score']
            entity.id = len(doc.ner_ents)
            entity.confidence = r['score']

            doc.ner_ents.append(entity)
        create_main_ann(doc)

    def _process(self,
                 stream: Iterable[Union[MutableDocument, None]],
                 batch_size_chars: int) -> Iterator[Optional[MutableDocument]]:
        if not hasattr(self, "ner_pipe"):
            self.create_eval_pipeline()
        for docs in self.batch_generator(
                stream, batch_size_chars):  # type: ignore
            # For now we will process the documents one by one, should be
            # improved in the future to use batching
            for doc in docs:
                self._process_doc(doc)
            yield from docs

    # Override
    def __call__(self, doc: MutableDocument) -> MutableDocument:
        """Process one document, used in the spacy pipeline for sequential
        document processing.

        Args:
            doc (Doc):
                A spacy document

        Returns:
            Doc: The same spacy document.
        """

        # Just call the pipe method
        doc = next(self.pipe(iter([doc])))

        return doc


# NOTE: Only needed for datasets backwards compatibility
def func_has_kwarg(func: Callable, keyword: str):
    sig = inspect.signature(func)
    return keyword in sig.parameters



import os
import shutil
import tempfile
from typing import Union, Literal

from medcat.storage.serialisables import Serialisable
from medcat.storage.serialisers import (
    serialise, deserialise, AvailableSerialisers)


def should_serialise_as_zip(path: str,
                            as_zip: Union[bool, Literal['auto']]
                            ) -> bool:
    if as_zip == 'auto':
        return path.endswith(".zip")
    return as_zip


def serialise_as_zip(
        obj: Serialisable, path: str,
        ser_type: Union[AvailableSerialisers, str] = AvailableSerialisers.dill,
        overwrite: bool = False,
        ) -> None:
    """Serialse the file to a .zip at the specified path.

    The process uses the regular `serialise` method to serialise the object
    as a folder into a temporary directory and subsequently zips it up to
    the path requested.

    Args:
        obj (Serialisable): The object to serialise.
        path (str): The path to serialse the file to. Should end with .zip.
        ser_type (Union[AvailableSerialisers, str], optional): The serialiser
            to use. Defaults to AvailableSerialisers.dill.
        overwrite (bool, optional):
            Whether to allow overwriting existing files. Defaults to False.
    """
    if not path.endswith('.zip'):
        raise ValueError(f"Path must end with .zip, got {path}")
    with tempfile.TemporaryDirectory() as tmpdir:
        serialise(ser_type, obj, tmpdir)
        if not overwrite and os.path.exists(path):
            raise ValueError(f"Cannot overwrite existing file: {path}")
        base_name = path[:-4]  # remove '.zip'
        shutil.make_archive(
            base_name=base_name,
            format='zip',
            root_dir=tmpdir,
        )


def deserialise_from_zip(path: str) -> Serialisable:
    """Deserialise from a zip file.

    The process involves unzipping the contents to a temporary directory,
    and subsequently using the ruglar `deserialise` method to deserialise
    from that.

    Args:
        path (str): The path to deserialise from. Should end with .zip.

    Returns:
        Serialisable: The deserialised object
    """
    if not path.endswith('.zip'):
        raise ValueError(f"Path must end with .zip, got {path}")
    with tempfile.TemporaryDirectory() as tmpdir:
        shutil.unpack_archive(path, tmpdir, format='zip')
        return deserialise(tmpdir)



from enum import Enum, auto
from typing import Union, Type, Any, Optional
import os
from abc import ABC, abstractmethod
from importlib import import_module
import logging
import re
import importlib

import dill as _dill

from medcat.storage.serialisables import Serialisable, ManualSerialisable
from medcat.storage.serialisables import get_all_serialisable_members
from medcat.storage.schema import load_schema, save_schema
from medcat.storage.schema import DEFAULT_SCHEMA_FILE, IllegalSchemaException
from medcat.utils.legacy.v2_beta import (
    fix_module_and_cls_name, RemappingUnpickler)


logger = logging.getLogger(__name__)


SER_TYPE_FILE = '.serialised_by'
MANUAL_SERIALISED_TAG = 'MANUALLY_SERIALISED:'
MANUAL_SERIALISED_RE = re.compile(re.escape(MANUAL_SERIALISED_TAG) + "(.*)")


class Serialiser(ABC):
    """The abstract serialiser base class.

    This class is responsible for both serialising and deserialising.
    """
    RAW_FILE = 'raw_dict.dat'

    @property
    @abstractmethod
    def ser_type(self) -> 'AvailableSerialisers':
        """The serialiser type."""
        pass

    @abstractmethod
    def serialise(self, raw_parts: dict[str, Any], target_file: str) -> None:
        """Serialise the raw attributes / objects.

        Args:
            raw_parts (dict[str, Any]): The raw objects to serialise.
            target_file (str): The file name to write to.
        """
        pass

    @abstractmethod
    def deserialise(self, target_file: str) -> dict[str, Any]:
        """Deserialise data written to the specified file.

        Args:
            target_file (str): The file to read from.

        Returns:
            dict[str, Any]: The deserialised raw attributes / objects.
        """
        pass

    @classmethod
    def get_ser_type_file(cls, folder: str) -> str:
        return os.path.join(folder, SER_TYPE_FILE)

    def save_ser_type_file(self, folder: str) -> None:
        """Save the serialiser type into the specified folder.

        Args:
            folder (str): The folder to use.
        """
        file_path = self.get_ser_type_file(folder)
        self.ser_type.write_to(file_path)

    @classmethod
    def get_manually_serialised_path(cls, folder: str) -> Optional[str]:
        file_path = cls.get_ser_type_file(folder)
        with open(file_path) as f:
            contents = f.read()
        matched = MANUAL_SERIALISED_RE.match(contents)
        if matched is not None:
            return matched.group(1)
        return None

    def check_ser_type(self, folder: str) -> None:
        """Check that the folder contains data serialised by this serialiser.

        Args:
            folder (str): Target folder.

        Raises:
            TypeError: If the folder was not serialised by this serialiser.
        """
        file_path = os.path.join(folder, SER_TYPE_FILE)
        in_folder = AvailableSerialisers.from_file(file_path)
        if in_folder != self.ser_type:
            raise TypeError(
                "Expected nested bits to be serialised by the same serialiser")

    def serialise_all(self, obj: Serialisable, target_folder: str,
                      overwrite: bool = False) -> None:
        """Serialise the entire object into the target folder.

        This finds the serialisable parts (attributes) of the object and calls
        the same method on them recursively.
        It also finds the raw attributes (if any) and serialises them.

        Args:
            obj (Serialisable): The object to serialise.
            target_folder (str): The target folder.
            overwrite (bool): Whether to allow overwriting. Defaults to False.

        Raises:
            IllegalSchemaException:
                If there's multiple parts with the same name or
                a file already exists.
        """
        if isinstance(obj, ManualSerialisable):
            obj_cls = type(obj)
            logger.info("Serialising obj '%s' manually", obj_cls.__name__)
            obj.serialise_to(target_folder)
            # write the serialised_by meta file
            cls_path = obj_cls.__module__ + "." + obj_cls.__name__
            with open(self.get_ser_type_file(target_folder), 'w') as f:
                f.write(MANUAL_SERIALISED_TAG + cls_path)
            return
        ser_parts, raw_parts = get_all_serialisable_members(obj)
        for part, name in ser_parts:
            basename = name
            part_folder = os.path.join(target_folder, basename)
            if os.path.exists(part_folder) and not overwrite:
                raise IllegalSchemaException(
                    f"File already exists: {part_folder}. Unable to overwrite")
            elif not os.path.exists(part_folder):
                os.mkdir(part_folder)
            # recursive
            self.serialise_all(part, part_folder, overwrite=overwrite)
        if raw_parts:
            raw_file = os.path.join(target_folder, self.RAW_FILE)
            self.serialise(raw_parts, raw_file)
        schema_path = os.path.join(target_folder, DEFAULT_SCHEMA_FILE)
        save_schema(schema_path, obj.__class__, obj.get_init_attrs())
        self.save_ser_type_file(target_folder)

    @classmethod
    def deserialise_manually(cls, folder_path: str, man_cls_path: str,
                             **init_kwargs) -> Serialisable:
        logger.info("Deserialising manually based on %s", man_cls_path)
        module_name, cls_name = man_cls_path.rsplit(".", 1)
        module_name, cls_name = fix_module_and_cls_name(module_name, cls_name)
        rel_module = importlib.import_module(module_name)
        man_cls = getattr(rel_module, cls_name)
        if not issubclass(man_cls, ManualSerialisable):
            raise ValueError(
                f"Cannot manually serialise {folder_path} "
                "because the class used for manual serialisation "
                f"({man_cls_path}) does not implement ManualSerialisable")
        return man_cls.deserialise_from(folder_path, **init_kwargs)

    def deserialise_all(self, folder_path: str,
                        ignore_folders_prefix: set[str] = set(),
                        ignore_folders_suffix: set[str] = set(),
                        **kwargs) -> Serialisable:
        """Deserialise contents of folder.

        Additional initialisation keyword arguments can be provided if needed.

        This loads both the raw attributes for this object as well as the
        serialisable parts / attributes recursively.

        Args:
            folder_path (str): The folder path.
            ignore_folders_prefix (set[str]): The prefixes of folders
                to ignore.
            ignore_folders_suffix (set[str]): The suffixes of folders
                to ignore.

        Returns:
            Serialisable: The resulting object.
        """
        man_cls_path = self.get_manually_serialised_path(folder_path)
        if man_cls_path:
            return self.deserialise_manually(folder_path, man_cls_path)
        self.check_ser_type(folder_path)
        schema_path = os.path.join(folder_path, DEFAULT_SCHEMA_FILE)
        cls_path, init_attrs = load_schema(schema_path)
        module_path, cls_name = cls_path.rsplit('.', 1)
        module = import_module(module_path)
        cls: Type = getattr(module, cls_name)
        init_kwargs: dict[str, Serialisable] = kwargs
        non_init_sers: dict[str, Serialisable] = {}
        for part_name in os.listdir(folder_path):
            if part_name == DEFAULT_SCHEMA_FILE or part_name == self.RAW_FILE:
                continue
            suitable_folder = True
            for ignore_prefix in ignore_folders_prefix:
                if part_name.startswith(ignore_prefix):
                    suitable_folder = False
                    break
            for ignore_suffix in ignore_folders_suffix:
                if part_name.endswith(ignore_suffix):
                    suitable_folder = False
                    break
            if not suitable_folder:
                continue
            part_path = os.path.join(folder_path, part_name)
            if not os.path.isdir(part_path):
                continue
            part = self.deserialise_all(
                part_path, ignore_folders_prefix=ignore_folders_prefix,
                ignore_folders_suffix=ignore_folders_suffix)
            if part_name in init_attrs:
                init_kwargs[part_name] = part
            else:
                non_init_sers[part_name] = part
        raw_file = os.path.join(folder_path, self.RAW_FILE)
        raw_parts: dict[str, Any]
        if os.path.exists(raw_file):
            raw_parts = self.deserialise(raw_file)
        else:
            raw_parts = {}
        missing = set(set(init_attrs) - set(init_kwargs))
        if init_attrs and missing:
            for missed in missing:
                init_kwargs[missed] = raw_parts.pop(missed)
        obj = cls(**init_kwargs)
        all_items = list(raw_parts.items()) + list(non_init_sers.items())
        for attr_name, attr in all_items:
            setattr(obj, attr_name, attr)
        return obj


class AvailableSerialisers(Enum):
    """Describes the available serialisers."""
    dill = auto()
    json = auto()

    def write_to(self, file_path: str) -> None:
        with open(file_path, 'w') as f:
            f.write(self.name)

    @classmethod
    def from_file(cls, file_path: str) -> 'AvailableSerialisers':
        with open(file_path, 'r') as f:
            return cls[f.read().strip()]


class DillSerialiser(Serialiser):
    """The dill based serialiser."""
    ser_type = AvailableSerialisers.dill

    def serialise(self, raw_parts: dict[str, Any], target_file: str) -> None:
        with open(target_file, 'wb') as f:
            _dill.dump(raw_parts, f)

    def deserialise(self, target_file: str) -> dict[str, Any]:
        with open(target_file, 'rb') as f:
            return RemappingUnpickler(f).load()


_DEF_SER = AvailableSerialisers.dill


def get_serialiser(
        serialiser_type: Union[str, AvailableSerialisers] = _DEF_SER
                   ) -> Serialiser:
    """Get the serialiser based on the type specified.

    Args:
        serialiser_type (Union[str, AvailableSerialisers], optional):
            The required type. Defaults to 'dill'.

    Raises:
        ValueError: If no serialiser is found.

    Returns:
        Serialiser: The appropriate serialiser.
    """
    if isinstance(serialiser_type, str):
        serialiser_type = AvailableSerialisers[serialiser_type.lower()]
    if serialiser_type is AvailableSerialisers.dill:
        return DillSerialiser()
    elif serialiser_type is AvailableSerialisers.json:
        from medcat.storage.jsonserialiser import JsonSerialiser
        return JsonSerialiser()
    raise ValueError("Unknown or unimplemented serialsier type: "
                     f"{serialiser_type}")


def get_serialiser_type_from_folder(folder_path: str) -> AvailableSerialisers:
    """Get the serialiser type that was used to serialise data in the folder.

    Args:
        folder_path (str): The folder in question.

    Returns:
        AvailableSerialisers: The serialiser type.
    """
    file_path = os.path.join(folder_path, SER_TYPE_FILE)
    return AvailableSerialisers.from_file(file_path)


def get_serialiser_from_folder(folder_path: str) -> Serialiser:
    """Get the serialiser that was used to serialise the data in the folder.

    Args:
        folder_path (str): The folder in question.

    Returns:
        Serialiser: The appropriate serialiser.
    """
    ser_type = get_serialiser_type_from_folder(folder_path)
    logger.info("Determined serialised of type %s off disk",
                ser_type.name)
    return get_serialiser(ser_type)


def serialise(serialiser_type: Union[str, AvailableSerialisers],
              obj: Serialisable, target_folder: str,
              overwrite: bool = False) -> None:
    """Serialise an object based on the specified serialiser type.

    Args:
        serialiser_type (Union[str, AvailableSerialisers]):
            The serialiser type.
        obj (Serialisable):
            The object to serialise.
        target_folder (str):
            The folder to serialise into.
        overwrite (bool):
            Whether to allow overwriting. Defaults to False.
    """
    ser = get_serialiser(serialiser_type)
    ser.serialise_all(obj, target_folder, overwrite=overwrite)


def deserialise(folder_path: str,
                ignore_folders_prefix: set[str] = set(),
                ignore_folders_suffix: set[str] = set(),
                **init_kwargs) -> Serialisable:
    """Deserialise contents of a folder.

    Extra init keyword arguments can be provided if needed.
    These are generally:
    - cnf: The config relevant to the components
    - tokenizer (BaseTokenizer): The base tokenizer for the model
    - cdb (CDB): The CDB for the model
    - vocab (Vocab): The Vocab for the model
    - model_load_path (Optional[str]): The model load path,
        but not the component load path

    This method finds the serialiser to be used based on the files on disk.

    Args:
        folder_path (str): The folder to serialise.
        ignore_folders_prefix (set[str]): The prefixes of folders to ignore.
        ignore_folders_suffix (set[str]): The suffixes of folders to ignore.

    Returns:
        Serialisable: The deserialised object.
    """
    # if manually serialised, do manually deserialisation
    man_cls_path = Serialiser.get_manually_serialised_path(folder_path)
    if man_cls_path:
        return Serialiser.deserialise_manually(folder_path, man_cls_path,
                                               **init_kwargs)
    ser = get_serialiser_from_folder(folder_path)
    return ser.deserialise_all(
        folder_path, ignore_folders_prefix=ignore_folders_prefix,
        ignore_folders_suffix=ignore_folders_suffix, **init_kwargs)



from typing import Optional
from contextlib import nullcontext

from medcat.config.config import LinkingFilters
from medcat.data.mctexport import MedCATTrainerExportProject
from medcat.utils.config_utils import temp_changed_config


def project_filters(filters: LinkingFilters,
                    project: MedCATTrainerExportProject,
                    extra_cui_filter: Optional[set[str]],
                    use_project_filters: bool):
    """Context manager with per project filters based on a trainer export.

    Args:
        filters (LinkingFilters): The current config.
        project (MedCATTrainerExportProject): The trainer export.
        extra_cui_filter (Optional[set[str]]): Extra cui filters.
        use_project_filters (bool): Whether to use project filters.
    """
    if extra_cui_filter is not None and not use_project_filters:
        return temp_changed_config(filters, 'cuis', extra_cui_filter)
    if use_project_filters:
        cuis = project.get('cuis', None)
        if cuis is None or not cuis:
            return nullcontext()
        return temp_changed_config(filters, 'cuis', set(cuis.split(",")))
    return temp_changed_config(filters, 'cuis', set())



import os
from typing import Optional
from transformers.models.bert.tokenization_bert_fast import BertTokenizerFast
from transformers import PreTrainedTokenizerFast
import logging

from medcat.config.config_rel_cat import ConfigRelCAT


logger = logging.getLogger(__name__)


class BaseTokenizerWrapper(PreTrainedTokenizerFast):

    name = "base_tokenizer_wrapper_rel"

    def __init__(self, hf_tokenizers=None,
                 max_seq_length: Optional[int] = None,
                 add_special_tokens: Optional[bool] = False):
        self.hf_tokenizers = hf_tokenizers
        self.max_seq_length = max_seq_length
        self._add_special_tokens = add_special_tokens

    def get_size(self):
        return len(self.hf_tokenizers.vocab)

    def token_to_id(self, token):
        return self.hf_tokenizers.convert_tokens_to_ids(token)

    def get_pad_id(self):
        return self.hf_tokenizers.pad_token_id

    def __call__(self, text, truncation: Optional[bool] = True):
        if isinstance(text, str):
            result = self.hf_tokenizers.encode_plus(
                text, return_offsets_mapping=True, return_length=True,
                return_token_type_ids=True, return_attention_mask=True,
                add_special_tokens=self._add_special_tokens,
                max_length=self.max_seq_length, padding="longest",
                truncation=truncation)

            return {'offset_mapping': result['offset_mapping'],
                    'input_ids': result['input_ids'],
                    'tokens':  self.hf_tokenizers.convert_ids_to_tokens(
                        result['input_ids']),
                    'token_type_ids': result['token_type_ids'],
                    'attention_mask': result['attention_mask'],
                    'length': result['length']
                    }
        elif isinstance(text, list):
            results = self.hf_tokenizers._batch_encode_plus(
                text, return_offsets_mapping=True, return_length=True,
                return_token_type_ids=True,
                add_special_tokens=self._add_special_tokens,
                max_length=self.max_seq_length, truncation=truncation)
            output = []
            for ind in range(len(results['input_ids'])):
                output.append({
                    'offset_mapping': results['offset_mapping'][ind],
                    'input_ids': results['input_ids'][ind],
                    'tokens':  self.hf_tokenizers.convert_ids_to_tokens(
                        results['input_ids'][ind]),
                    'token_type_ids': results['token_type_ids'][ind],
                    'attention_mask': results['attention_mask'][ind],
                    'length': result['length']
                })
            return output
        else:
            raise Exception(
                "Unsupported input type, supported: text/list, but got: "
                f"{type(text)}")

    def save(self, dir_path: str):
        path = os.path.join(dir_path, self.name)
        self.hf_tokenizers.save_pretrained(path)

    @classmethod
    def load(cls, tokenizer_path: str, relcat_config: ConfigRelCAT, **kwargs
             ) -> "BaseTokenizerWrapper":

        tokenizer = BaseTokenizerWrapper()

        cnf_gen = relcat_config.general

        if os.path.exists(tokenizer_path):
            if "modern-bert" in cnf_gen.tokenizer_name:
                from medcat.components.addons.relation_extraction.modernbert.tokenizer import TokenizerWrapperModernBERT  # noqa
                tokenizer = TokenizerWrapperModernBERT.load(
                    tokenizer_path, relcat_config=relcat_config, **kwargs)
            elif "bert" in cnf_gen.tokenizer_name:
                from medcat.components.addons.relation_extraction.bert.tokenizer import TokenizerWrapperBERT  # noqa
                tokenizer = TokenizerWrapperBERT.load(
                    tokenizer_path, relcat_config=relcat_config, **kwargs)
            elif "llama" in cnf_gen.tokenizer_name:
                from medcat.components.addons.relation_extraction.llama.tokenizer import TokenizerWrapperLlama  # noqa
                tokenizer = TokenizerWrapperLlama.load(
                    tokenizer_path, relcat_config=relcat_config, **kwargs)
            logger.info("Tokenizer loaded %s from: %s",
                        str(tokenizer.__class__.__name__), tokenizer_path)
        elif cnf_gen.model_name:
            logger.info("Attempted to load Tokenizer from path: %s,"
                        " but it doesn't exist, loading default toknizer from "
                        "model_name relcat_config.general.model_name: %s",
                        tokenizer_path, cnf_gen.model_name)
            from medcat.components.addons.relation_extraction.bert.tokenizer import TokenizerWrapperBERT  # noqa
            from medcat.components.addons.relation_extraction.ml_utils import create_tokenizer_pretrain  # noqa
            logger.info(
                "Addeding special tokens to tokenizer: %s %s",
                str(cnf_gen.tokenizer_relation_annotation_special_tokens_tags),
                str(cnf_gen.tokenizer_other_special_tokens))
            tokenizer = TokenizerWrapperBERT(
                BertTokenizerFast.from_pretrained(cnf_gen.model_name),
                add_special_tokens=True)
            tokenizer = create_tokenizer_pretrain(
                tokenizer, relcat_config=relcat_config)
        else:
            logger.info(
                "Attempted to load Tokenizer from path: %s, "
                "but it doesn't exist, loading default toknizer from "
                "model_name config.general.model_name:bert-base-uncased",
                tokenizer_path)
            from medcat.components.addons.relation_extraction.bert.tokenizer import TokenizerWrapperBERT  # noqa
            tokenizer = TokenizerWrapperBERT(
                BertTokenizerFast.from_pretrained(cnf_gen.model_name),
                max_seq_length=cnf_gen.max_seq_length,
                add_special_tokens=cnf_gen.tokenizer_special_tokens)
        return tokenizer



import os
import logging
from typing import Any, Union, Optional

from medcat.config.config import SerialisableBaseModel, ComponentConfig
from medcat.storage.serialisers import deserialise


class General(SerialisableBaseModel):
    """The General part of the RelCAT config"""
    device: str = "cpu"
    """The device to use (CPU or GPU).

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    relation_type_filter_pairs: list = []
    """Map from category values to ID, if empty it will be
    autocalculated during training"""
    vocab_size: Optional[int] = None
    lowercase: bool = True
    """If true all input text will be lowercased"""
    cntx_left: int = 15
    """Number of tokens to take from the left of the concept"""
    cntx_right: int = 15
    """Number of tokens to take from the right of the concept"""
    window_size: int = 300
    """Max acceptable dinstance between entities (in characters),
    care when using this as it can produce sentences that
    are over 512 tokens (limit is given by tokenizer)"""

    limit_samples_per_class: int = -1
    """Number of samples per class, this limit is applied for train samples,
    o if train samples are 100 then test would be 20."""
    addl_rels_max_sample_size: int = 200
    """Limit the number of 'Other' samples selected for training/test. This is
    applied per encountered medcat project, sample_size/num_projects. """
    create_addl_rels: bool = False
    """When processing relations from a MedCAT export/docs, relations labeled
    as 'Other' are created from all the annotations pairs available"""
    create_addl_rels_by_type: bool = False
    """When creating the 'Other' relation class, actually split this class
    into subclasses based on concept types"""

    tokenizer_name: str = "bert"
    """The name of the tokenizer user.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    model_name: str = "bert-base-uncased"
    """The name of the model used.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    log_level: int = logging.INFO
    """The log level for RelCAT.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    max_seq_length: int = 512
    """The maximum sequence length.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    tokenizer_special_tokens: bool = False
    """Tokenizer.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    annotation_schema_tag_ids: list[int] = [30522, 30523, 30524, 30525]
    """If a foreign non-MCAT trainer dataset is used, you can insert your own
    Rel entity token delimiters into the tokenizer, copy those token IDs here,
    and also resize your tokenizer embeddings and adjust the hidden_size of
    the model, this will depend on the number of tokens you introduce for
    example: 30522 - [s1], 30523 - [e1], 30524 - [s2], 30525 - [e2],
    30526 - [BLANK], 30527 - [ENT1], 30528 - [ENT2], 30529 - [/ENT1],
    30530 - [/ENT2]
    Please note that the tokenizer special tokens are supposed to be in pairs
    of two for example [s1] and [e1], [s2] and [e2], the [BLANK] is just an
    example placeholder token. If you have more than four tokens here then you
    need to make sure they are present in the text,  otherwise the pipeline
    will throw an error in the get_annotation_schema_tag() function.
    """

    tokenizer_relation_annotation_special_tokens_tags: list[str] = [
        "[s1]", "[e1]", "[s2]", "[e2]"]

    tokenizer_other_special_tokens: dict[str, str] = {"pad_token": "[PAD]"}
    """
    The special tokens used by the tokenizer. The {PAD} is
    for Lllama tokenizer."""

    labels2idx: dict[str, int] = {}
    idx2labels: dict[int, str] = {}

    pin_memory: bool = True
    """If True the data loader will copy the tensors to the GPU pinned memory.
    """

    seed: int = 13
    """The seed for random number generation.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    task: str = "train"
    """The task for RelCAT."""

    language: str = "en"
    """Used for Spacy lang setting"""

    @classmethod
    def convert_keys_to_int(cls, value):
        if isinstance(value, dict):
            return {int(k): v for k, v in value.items()}
        return value

    def __setattr__(self, key: str, value: Any):
        if key == "idx2labels" and isinstance(value, dict):
            value = self.convert_keys_to_int(value)  # Ensure conversion
        super().__setattr__(key, value)

    class Config:
        protected_namespaces = ()


class Model(SerialisableBaseModel):
    """The model part of the RelCAT config"""
    input_size: int = 300
    hidden_size: int = 768
    """The hidden size.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    hidden_layers: int = 3
    """ hidden_size * 5, 5 being the number of tokens,
    default (s1,s2,e1,e2+CLS).

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    model_size: int = 5120
    """The size of the model.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    dropout: float = 0.2
    num_directions: int = 2
    """2 - bidirectional model, 1 - unidirectional"""

    freeze_layers: bool = True
    """If we update the weights during training"""

    padding_idx: int = -1
    emb_grad: bool = True
    """If True the embeddings will also be trained"""
    ignore_cpos: bool = False
    """If set to True center positions will be ignored when calculating
    representation"""

    llama_use_pooled_output: bool = False
    """If set to True, used only in Llama model, it will add the extra tensor
    formed from selecting the max of the last hidden layer"""

    class Config:
        extra = 'allow'
        validate_assignment = True
        protected_namespaces = ()


class Train(SerialisableBaseModel):
    """The train part of the RelCAT config"""
    nclasses: int = 2
    """Number of classes that this model will output"""
    batch_size: int = 25
    """batch size"""
    nepochs: int = 1
    """Epochs"""
    lr: float = 1e-4
    """Learning rate"""
    stratified_batching: bool = False
    """Train the model with stratified batching"""
    batching_samples_per_class: list = []
    """Number of samples per class in each batch
    example for batch size 64: [6,6,6,8,8,8,6,8,8]"""
    batching_minority_limit: Union[list[int], int] = 0
    """Maximum number of samples the minority class can have.
    Since the minority class elements need to be repeated, this is used to
    facilitate that
    example: batching_samples_per_class - [6,6,6,8,8,8,6,8,8]
             batching_minority_limit - 6"""
    adam_betas: tuple[float, float] = (0.9, 0.999)
    adam_weight_decay: float = 0
    adam_epsilon: float = 1e-8
    test_size: float = 0.2
    gradient_acc_steps: int = 1
    multistep_milestones: list[int] = [
        2, 4, 6, 8, 12, 15, 18, 20, 22, 24, 26, 30]
    multistep_lr_gamma: float = 0.8
    max_grad_norm: float = 1.0
    shuffle_data: bool = True
    """Used only during training, if set the dataset will be shuffled before
    train/test split"""
    class_weights: Union[list[float], None] = None
    enable_class_weights: bool = False
    score_average: str = "weighted"
    """What to use for averaging F1/P/R across labels"""
    auto_save_model: bool = True
    """Should the model be saved during training for best results"""

    class Config:
        extra = 'allow'
        validate_assignment = True


class ConfigRelCAT(ComponentConfig):
    """The RelCAT part of the config"""
    general: General = General()
    model: Model = Model()
    train: Train = Train()

    class Config:
        extra = 'allow'
        validate_assignment = True

    @classmethod
    def load(cls, load_path: str = "./") -> "ConfigRelCAT":
        """Load the config from a file.

        Args:
            load_path (str): Path to RelCAT config. Defaults to "./".

        Returns:
            ConfigRelCAT: The loaded config.
        """
        config = cls()
        if os.path.exists(load_path):
            if os.path.sep + "config" not in load_path:
                load_path = os.path.join(load_path, "config")
            config: ConfigRelCAT = deserialise(load_path)  # type: ignore
            logging.info("Loaded config.json")

        return config



import platform
import logging
import importlib.metadata
import re

from pydantic import BaseModel

from medcat.storage.serialisables import AbstractSerialisable


logger = logging.getLogger(__name__)


DEP_NAME_PATTERN = re.compile(r'^[a-zA-Z0-9\-_]+')


def get_direct_dependencies(include_extras: bool) -> list[str]:
    """Gets the direct dependencies of the current package and their versions.

    Args:
        include_extras (bool): Whether to include extras (like spacy).
    """
    # NOTE: __package__ would be medcat.utils in this case
    package = __package__.split('.', 1)[0]
    reqs = importlib.metadata.requires(package)
    if reqs is None:
        raise ValueError("Unable to find package direct dependencies")
    # filter out extras
    if not include_extras:
        reqs = [req for req in reqs
                if "; extra ==" not in req]
    # only keep name, not version
    # NOTE: all correct dependency names will match this regex
    reqs = [DEP_NAME_PATTERN.match(req).group(0).lower()  # type: ignore
            for req in reqs]
    return reqs


def get_transitive_deps(direct_deps: list[str]) -> dict[str, str]:
    """Get the transitive dependencies of the direct dependencies.

    Args:
        direct_deps (list[str]): List of direct dependencies.

    Returns:
        dict[str, str]: The dependency names and their corresponding versions.
    """
    all_deps: dict[str, str] = {}
    to_process = set(direct_deps)
    processed = set()
    # list installed packages for ease of use
    installed_packages = {
        dist.metadata['name'].lower()
        for dist in importlib.metadata.distributions()}

    while to_process:
        package = to_process.pop()
        if package in processed:
            continue

        processed.add(package)

        try:
            dist = importlib.metadata.distribution(package)
        except importlib.metadata.PackageNotFoundError:
            # NOTE: if not installed, we won't bother
            #       after all, if we can save the model, clearly
            #       everything is working
            continue
        requires = dist.requires or []

        for req in requires:
            match = DEP_NAME_PATTERN.match(req)
            if match is None:
                raise ValueError(f"Malformed dependency: {req}")
            dep_name = match.group(0).lower()
            if (dep_name and dep_name not in processed and
                    dep_name in installed_packages):
                all_deps[dep_name] = importlib.metadata.distribution(
                    dep_name).version
                to_process.add(dep_name)

    for direct in direct_deps:
        # remove direct dependencies if they were added
        all_deps.pop(direct, None)
    return all_deps


def get_installed_dependencies(include_extras: bool) -> dict[str, str]:
    """Get the installed packages and their versions.

    Args:
        include_extras (bool): Whether to include extras (like spacy).

    Returns:
        dict[str, str]: All installed packages and their versions.
    """
    direct_deps = get_direct_dependencies(include_extras)
    installed_packages: dict[str, str] = {}
    for package in importlib.metadata.distributions():
        req_name = package.metadata["name"].lower()
        # NOTE: we're checking against the '-' typed package name not
        #       the import name (which will have _ instead)
        req_name_dashes = req_name.replace("_", "-")
        if all(cn not in direct_deps for cn in
               [req_name, req_name_dashes]):
            continue
        installed_packages[req_name] = package.version
    return installed_packages


def is_dependency_installed(dependency: str) -> bool:
    """Checks whether a dependency is installed.

    This takes into account changes such as '-' vs '_'.
    For example, `typing-extensions` is a direct dependency,
    but its module path will be `typing_extension` and that's
    how we can find it as an installed dependency.

    Args:
        dependency (str): The dependency in question.

    Returns:
        bool: Whether the depedency has been installed.
    """
    installed_deps = get_installed_dependencies(True)
    dep_name = dependency.lower()
    dep_name_underscores = dependency.replace("-", "_")
    options = [dep_name, dep_name_underscores]
    return any(option in installed_deps for option in options)


class Environment(BaseModel, AbstractSerialisable):
    dependencies: dict[str, str]
    transitive_deps: dict[str, str]
    os: str
    cpu_arcitecture: str
    python_version: str

    @classmethod
    def get_init_attrs(cls) -> list[str]:
        return list(cls.model_fields)


def get_environment_info(include_transitive_deps: bool = True,
                         include_extras: bool = True) -> Environment:
    """Get the current environment information.

    This includes dependency versions, the OS, the CPU architecture and the
        python version.

    Args:
        include_transitive_deps (bool): Whether to include transitive
            dependencies. Defaults to True.
        include_extras (bool): Whether to include extras (like spacy).
            Defaults to True.

    Returns:
        Environment: The environment.
    """
    deps = get_installed_dependencies(include_extras)
    os = platform.platform()
    cpu_arc = platform.machine()
    py_ver = platform.python_version()
    if include_transitive_deps:
        direct_deps = list(deps.keys())
        trans_deps = get_transitive_deps(direct_deps)
    else:
        trans_deps = {}
    return Environment(dependencies=deps, transitive_deps=trans_deps, os=os,
                       cpu_arcitecture=cpu_arc, python_version=py_ver)



from sklearn.metrics import classification_report
import numpy as np
import pandas as pd
from collections import defaultdict
from scipy.special import softmax
import logging


logger = logging.getLogger(__name__)


def metrics(p, return_df=False, plus_recall=0, tokenizer=None, dataset=None,
            merged_negative={0, 1, -100}, padding_label=-100, csize=15,
            subword_label=1,
            verbose=False):
    """
    Calculate metrics for a model's predictions, based off the tokenized
    output of a MedCATTrainer project.

    Args:
        p: The model's predictions.
        return_df: Whether to return a DataFrame of metrics.
        plus_recall: The recall to add to the model's predictions.
        tokenizer: The tokenizer used to tokenize the texts.
        dataset: The dataset used to train the model.
        merged_negative: The negative labels to merge.
        padding_label: The padding label.
        csize: The size of the context window.
        subword_label: The subword label.
        verbose: Whether to print the metrics.
    Returns:
        Dict: A dictionary of metrics.
    """
    predictions = np.array(p.predictions)
    predictions = softmax(predictions, axis=2)
    examples = None
    if plus_recall > 0:
        # Devalue 0 and 1
        predictions[:, :, 0] = predictions[:, :, 0] - (
            predictions[:, :, 0] * plus_recall)
        predictions[:, :, 1] = predictions[:, :, 1] - (
            predictions[:, :, 1] * plus_recall)

    preds = np.argmax(predictions, axis=2)
    # Ignore predictions where label == -100, padding
    preds[np.where(p.label_ids == -100)] = -100

    if dataset is not None and tokenizer is not None:
        examples = {'fp': defaultdict(list), 'fn': defaultdict(list)}
        ilbl = {v: k for k, v in tokenizer.label_map.items()}
        for i in range(len(preds)):
            st = None
            for j in range(len(preds[i])):
                _p = preds[i][j]
                _l = p.label_ids[i][j]
                if (len(p.label_ids[i]) > (j + 1) and
                        p.label_ids[i][j + 1] != padding_label):
                    _p2 = preds[i][j + 1]
                    _l2 = p.label_ids[i][j + 1]
                else:
                    _p2 = None
                    _l2 = None

                _d = dataset[i]['input_ids']
                id = dataset[i]['id']
                name = dataset[i]['name']
                if _l not in {subword_label, padding_label}:
                    # We ignore padding and subwords
                    if _l != _p:
                        if st is None:
                            st = max(0, j - csize)
                            _j = j

                        if not (_l2 is not None and
                                _l2 == _l and _l2 != _p2 and _p2 == _p):
                            # We want to merge tokens if it is the same label
                            # and same prediction when recording the examples,
                            # that is why we have the if
                            dc1 = tokenizer.hf_tokenizer.decode(_d[st:_j])
                            dc2 = tokenizer.hf_tokenizer.decode(_d[_j:j + 1])
                            dc3 = tokenizer.hf_tokenizer.decode(
                                _d[j + 1:j + csize])
                            t = (dc1 + "<<" + str(dc2) + ">>" + dc3)
                            value = str(dc2).strip()
                            examples['fp'][ilbl[_p]].append(
                                ({'id': id, 'name': name, 'value': value,
                                  'label': tokenizer.cui2name.get(ilbl[_l],
                                                                  ilbl[_l]),
                                  'text': t}))
                            examples['fn'][ilbl[_l]].append(
                                ({'id': id, 'name': name, 'value': value,
                                  'prediction': tokenizer.cui2name.get(
                                      ilbl[_p], ilbl[_p]),
                                  'text': t}))

                            st = None

    _labels = np.reshape(p.label_ids, -1)
    _preds = np.reshape(preds, -1)

    cr = classification_report(_labels, _preds, output_dict=True)
    _cr = {}
    ignore = [str(tokenizer.label_map['O']), str(tokenizer.label_map['X']),
              '-100']
    ilabel_map = {str(v): k for k, v in tokenizer.label_map.items()}
    for key in cr.keys():
        if key not in ignore and key in ilabel_map:
            _cr[key] = cr[key]

    # Get merged metrics, ie all PID is just one entity
    tp = defaultdict(int)
    fn = defaultdict(int)
    tp_all = 0
    fp_all = 0
    for i, _label in enumerate(_labels):
        _pred = _preds[i]
        if _label in merged_negative:
            if _pred in merged_negative:
                tp[_label] += 1
            else:
                fn[_label] += 1

            if _label == 0:
                if _pred not in merged_negative:
                    fp_all += 1
        else:
            if _pred not in merged_negative:
                tp[_label] += 1
                tp_all += 1
            else:
                fn[_label] += 1
    for key in _cr:
        key = int(key)
        if int(key) in tp:
            _cr[str(key)]['r_merged'] = (
                tp[key] / (tp[key] + fn.get(key, 0))
                if tp[key] + fn.get(key, 0) > 0
                else 0)
        else:
            _cr[str(key)]['r_merged'] = None

    data = [['cui', 'name', 'p', 'r', 'f1', 'support', 'r_merged', 'p_merged']]
    for key in _cr:
        cui = ilabel_map[key]
        p_merged = tp_all / (tp_all + fp_all) if (tp_all + fp_all) > 0 else 0
        data.append(
            [cui, tokenizer.cui2name.get(cui, cui), _cr[key]['precision'],
             _cr[key]['recall'], _cr[key]['f1-score'], _cr[key]['support'],
             _cr[key]['r_merged'], p_merged])

    df = pd.DataFrame(data[1:], columns=data[0])
    if verbose:
        logger.info('%s', df)

    if not return_df:
        return {
            'recall': np.average(df.r.values),
            'precision': np.average(df.p.values),
            'f1': np.average(df.f1.values),
            'recall_merged': np.average([x for x in df.r_merged.values
                                         if pd.notna(x)]),
            'precison_merged': np.average([x for x in df.p_merged.values
                                           if pd.notna(x)])}
    else:
        return df, examples


def _anno_within_pred_list(label: dict, preds: list[dict]) -> bool:
    """
    Check if a label is within a list of predictions,

    Args:
        label (Dict): an annotation likely from a MedCATTrainer project
        preds (List[Dict]): a list of predictions likely from a cat.__call__
    Returns:
        bool: True if the label is within the list of predictions,
            False otherwise
    """
    return any(label['start'] >= p['start'] and label['end'] <= p['end']
               for p in preds)


def evaluate_predictions(true_annotations: list[list[dict]],
                         all_preds: list[list[dict]], texts: list[str],
                         cui2preferred_name: dict[str, str]
                         ) -> tuple[pd.DataFrame, dict]:
    """
    Evaluate predictions against sets of collected labels as collected and
    utput from a MedCATTrainer project.
    Counts predictions as correct if the prediction fully encloses the label.

    Args:
        true_annotations (list[list[dict]]): Ground truth predictions by text
        all_preds (list[list[dict]]): Model predictions by text
        texts (list[str]): Original list of texts
        cui2preferred_name (dict[str, str]): Dictionary of CUI to preferred
            name, likely to be cat.cdb.cui2preferred_name.
    Returns:
        tuple[pd.DataFrame, Dict]: A tuple containing a DataFrame of
            evaluation metrics and a dictionary of missed annotations per CUI.
    """
    per_cui_recall = {}
    per_cui_prec = {}
    per_cui_recall_merged = {}
    per_cui_anno_counts = {}
    per_cui_annos_missed = defaultdict(list)
    uniq_labels = set([p['cui'] for ap in true_annotations for p in ap])

    for cui in uniq_labels:
        # annos in test set
        anno_count = sum([len([p for p in cui_annos if p['cui'] == cui])
                          for cui_annos in true_annotations])
        pred_counts = sum([len([p for p in d if p['cui'] == cui])
                           for d in all_preds])

        # print(anno_count)
        # print(pred_counts)

        # print(f'pred_count: {pred_counts}, anno_count:{anno_count}')
        per_cui_anno_counts[cui] = anno_count

        doc_annos_left, preds_left, doc_annos_left_any_cui = [], [], []

        for doc_preds, doc_labels, text in zip(
                all_preds, true_annotations, texts):
            # num of annos that are not found - recall
            cui_labels = [label for label in doc_labels if label['cui'] == cui]
            cui_doc_preds = [pred for pred in doc_preds if pred['cui'] == cui]

            labels_not_found = [
                label for label in cui_labels
                if not _anno_within_pred_list(label, cui_doc_preds)]
            doc_annos_left.append(len(labels_not_found))

            # num of annos that are not found across any cui prediction
            # - recall_merged
            any_labels_not_found = [
                label for label in cui_labels
                if not _anno_within_pred_list(label, doc_preds)]
            doc_annos_left_any_cui.append(len(any_labels_not_found))

            per_cui_annos_missed[cui].append(any_labels_not_found)

            # num of preds that are incorrect - precision
            preds_left.append(len([
                label for label in cui_doc_preds
                if not _anno_within_pred_list(label, cui_labels)]))

        if anno_count != 0 and pred_counts != 0:
            per_cui_recall[cui] = (
                anno_count - sum(doc_annos_left)) / anno_count
            per_cui_recall_merged[cui] = (
                anno_count - sum(doc_annos_left_any_cui)) / anno_count
            per_cui_prec[cui] = (
                pred_counts - sum(preds_left)) / pred_counts
        else:
            per_cui_recall[cui] = 0
            per_cui_recall_merged[cui] = 0
            per_cui_prec[cui] = 0

    res_df = pd.DataFrame({
        'cui': per_cui_recall_merged.keys(),
        'recall_merged': per_cui_recall_merged.values(),
        'recall': per_cui_recall.values(),
        'precision': per_cui_prec.values(),
        'label_count': per_cui_anno_counts.values()},
        index=[cui2preferred_name[k] for k in per_cui_recall_merged])

    return res_df, per_cui_annos_missed



from typing import Optional

import logging
from medcat.tokenizing.tokens import MutableDocument
from medcat.components.types import CoreComponentType, AbstractCoreComponent
from medcat.components.ner.vocab_based_annotator import maybe_annotate_name
from medcat.tokenizing.tokenizers import BaseTokenizer
from medcat.vocab import Vocab
from medcat.cdb import CDB
from medcat.config.config import ComponentConfig


logger = logging.getLogger(__name__)


class NER(AbstractCoreComponent):
    name = 'cat_ner'

    def __init__(self, tokenizer: BaseTokenizer,
                 cdb: CDB) -> None:
        self.tokenizer = tokenizer
        self.cdb = cdb
        self.config = self.cdb.config

    def get_type(self) -> CoreComponentType:
        return CoreComponentType.ner

    def __call__(self, doc: MutableDocument) -> MutableDocument:
        """Detect candidates for concepts - linker will then be able
        to do the rest. It adds `entities` to the doc.entities and each
        entity can have the entity.link_candidates - that the linker
        will resolve.

        Args:
            doc (MutableDocument):
                Spacy document to be annotated with named entities.

        Returns:
            doc (MutableDocument):
                Spacy document with detected entities.
        """
        max_skip_tokens = self.config.components.ner.max_skip_tokens
        _sep = self.config.general.separator
        # Just take the tokens we need
        _doc = [tkn for tkn in doc if not tkn.to_skip]
        for i, tkn in enumerate(_doc):
            tkn = _doc[i]
            tkns = [tkn]
            # name_versions = [tkn.lower_, tkn._.norm]
            # name_versions = [tkn.norm, tkn.base.lower]
            name_versions = tkn.base.text_versions
            name = ""

            for name_version in name_versions:
                if self.cdb.has_subname(name_version):
                    if name:
                        name = name + _sep + name_version
                    else:
                        name = name_version
                    break
            # if name is in CDB
            if name in self.cdb.name2info and not tkn.base.is_stop:
                maybe_annotate_name(self.tokenizer, name, tkns, doc,
                                    self.cdb, self.config)
            # if name is not a subname CDB (explicitly)
            if not name:
                # There has to be at least something appended to the name
                # to go forward
                continue
            # if name is a part of a concept
            # we start adding onto it to get a match
            for j in range(i + 1, len(_doc)):
                if (_doc[j].base.index - _doc[j - 1].base.index - 1
                        > max_skip_tokens):
                    # Do not allow to skip more than limit
                    break
                tkn = _doc[j]
                tkns.append(tkn)
                # name_versions = [tkn.norm, tkn.base.lower]
                name_versions = tkn.base.text_versions

                name_changed = False
                name_reverse = None
                for name_version in name_versions:
                    _name = name + _sep + name_version
                    if self.cdb.has_subname(_name):
                        # Append the name and break
                        name = _name
                        name_changed = True
                        break

                    if self.config.components.ner.try_reverse_word_order:
                        _name_reverse = name_version + _sep + name
                        if self.cdb.has_subname(_name_reverse):
                            # Append the name and break
                            name_reverse = _name_reverse

                if name_changed:
                    if name in self.cdb.name2info:
                        maybe_annotate_name(self.tokenizer, name, tkns, doc,
                                            self.cdb, self.config)
                elif name_reverse is not None:
                    if name_reverse in self.cdb.name2info:
                        maybe_annotate_name(self.tokenizer, name_reverse, tkns,
                                            doc, self.cdb, self.config)
                else:
                    break

        return doc

    @classmethod
    def create_new_component(
            cls, cnf: ComponentConfig, tokenizer: BaseTokenizer,
            cdb: CDB, vocab: Vocab, model_load_path: Optional[str]) -> 'NER':
        return cls(tokenizer, cdb)



from medcat.tokenizing.tokenizers import MutableDocument, MutableEntity


# NOTE: the following used (in medcat v1) check tuis
#       but they were never passed to the method so
#       I've omitted it now
def create_main_ann(doc: MutableDocument) -> None:
    """Creates annotation in the spacy ents list
    from all the annotations for this document.

    Args:
        doc (Doc): Spacy document.
    """
    doc.ner_ents.sort(key=lambda x: len(x.base.text), reverse=True)
    tkns_in = set()
    main_anns: list[MutableEntity] = []
    for ent in doc.ner_ents:
        to_add = True
        for tkn in ent:
            if tkn in tkns_in:
                to_add = False
        if to_add:
            for tkn in ent:
                tkns_in.add(tkn)
            main_anns.append(ent)
    doc.linked_ents = sorted(list(doc.linked_ents) + main_anns,  # type: ignore
                            key=lambda ent: ent.base.start_char_index)



import dill
import logging

from medcat.cdb import CDB
from medcat.config import Config
from medcat.cdb.concepts import get_new_cui_info, get_new_name_info, TypeInfo
from medcat.utils.legacy.convert_config import get_config_from_nested_dict


logger = logging.getLogger(__name__)


class LegacyClassNotFound:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        return f"<LegacyClassNotFound args={self.args} kwargs={self.kwargs}>"


class CustomUnpickler(dill.Unpickler):
    def find_class(self, module, name):
        try:
            return super().find_class(module, name)
        except (AttributeError, ModuleNotFoundError):
            logger.warning(
                "Missing class %s.%s, replacing with LegacyClassNotFound.",
                module, name)
            return LegacyClassNotFound


def load_old_raw_data(old_path: str) -> dict:
    """Looads the raw data from old file.

    This uses a wrapper that allows loading the data even if the classes
    do not exist.

    Args:
        old_path (str): The path of the file to read.

    Returns:
        dict: The resulting raw data.
    """
    with open(old_path, 'rb') as f:
        # NOTE: custom unpickler needed because we
        #       do not have access to original modules within medcat(v1)
        data = CustomUnpickler(f).load()
    return data


EXPECTED_USEFUL_KEYS = [
    'name2cuis', 'name2cuis2status', 'name2count_train', 'name_isupper',

    'snames',

    'cui2names', 'cui2snames', 'cui2context_vectors', 'cui2count_train',
    'cui2tags', 'cui2type_ids', 'cui2preferred_name', 'cui2average_confidence',

    'addl_info',
    'vocab',
]
NAME2KEYS = {'name2cuis', 'name2cuis2status', 'name2count_train',
             'name_isupper'}
OPTIONAL_NAME2_KEYS = {"name_isupper", }
CUI2KEYS = {'cui2names', 'cui2snames', 'cui2context_vectors',
            'cui2count_train', 'cui2info', 'cui2tags', 'cui2type_ids',
            'cui2preferred_name', 'cui2average_confidence', }
CUI2KEYS_OPTIONAL = {"cui2info", }
TO_RENAME = {'vocab': 'token_counts'}


_DEFAULT_SNOMED_TYPE_ID2NAME = {
    '32816260': 'physical object', '2680757': 'observable entity',
    '37552161': 'body structure', '91776366': 'product',
    '81102976': 'organism', '28321150': 'procedure',
    '67667581': 'finding', '7882689': 'qualifier value',
    '91187746': 'substance', '29422548': 'core metadata concept',
    '40357424': 'foundation metadata concept',
    '33782986': 'morphologic abnormality', '9090192': 'disorder',
    '90170645': 'record artifact', '66527446': 'cell structure',
    '3061879': 'situation', '16939031': 'occupation',
    '31601201': 'person', '37785117': 'medicinal product',
    '17030977': 'assessment scale', '47503797': 'regime/therapy',
    '33797723': 'event', '82417248': 'navigational concept',
    '75168589': 'environment', '9593000': 'medicinal product form',
    '99220404': 'cell', '13371933': 'social concept',
    '46922199': 'religion/philosophy', '27603525': 'clinical drug',
    '43039974': 'attribute', '43857361': 'physical force',
    '40584095': 'metadata', '337250': 'specimen',
    '46506674': 'disposition', '87776218': 'role',
    '30703196': 'tumor staging', '31685163': 'staging scale',
    '21114934': 'dose form', '70426313': 'namespace concept',
    '51120815': 'intended site', '45958968': 'administration method',
    '51885115': 'OWL metadata concept', '8067332': 'basic dose form',
    '95475658': 'product name', '43744943': 'supplier',
    '66203715': 'transformation', '64755083': 'release characteristic',
    '49144999': 'state of matter', '39041339': 'unit of presentation',
    '18854038': 'geographic location', '28695783': 'link assertion',
    '14654508': 'racial group', '20410104': 'ethnic group',
    '92873870': 'special concept', '72706784': 'product',
    '78096516': 'environment / location', '25624495': 'SNOMED RT+CTV3',
    '55540447': 'linkage concept', '3242456': 'social context'
}


def _add_cui_info(cdb: CDB, data: dict) -> CDB:
    all_cuis = set()
    for key in CUI2KEYS:
        if key in CUI2KEYS_OPTIONAL and key not in data:
            logger.info(
                "Optional key to be converted was not found in target "
                "data: %s. Ignoring", key)
            continue
        ccuis = data[key].keys()
        logger.debug("Adding %d cuis based on '%s", len(ccuis), key)
        all_cuis.update(ccuis)
    logger.info("A total of %d CUIs identified", len(all_cuis))
    cui2names, cui2snames = data['cui2names'], data['cui2snames']
    cui2cv, cui2ct = data['cui2context_vectors'], data['cui2count_train']
    cui2tags, cui2type_ids = data['cui2tags'], data['cui2type_ids']
    cui2prefname = data['cui2preferred_name']
    cui2av_conf = data['cui2average_confidence']
    for cui in all_cuis:
        names = cui2names.get(cui, set())
        snames = cui2snames.get(cui, set())
        vecs = cui2cv.get(cui, None)
        count_train = cui2ct.get(cui, 0)
        tags = cui2tags.get(cui, None)
        type_ids = cui2type_ids.get(cui, set())
        prefname = cui2prefname.get(cui, None)
        av_conf = cui2av_conf.get(cui, 0.0)
        info = get_new_cui_info(
            cui=cui, preferred_name=prefname, names=names, subnames=snames,
            type_ids=type_ids, tags=tags, count_train=count_train,
            context_vectors=vecs, average_confidence=av_conf,
        )
        cdb.cui2info[cui] = info
    all_cui_tuis = set((ci['cui'], tui) for ci in cdb.cui2info.values()
                       for tui in ci['type_ids'])
    all_tuis = set(tui for _, tui in all_cui_tuis)
    if (not (tid2name := data['addl_info'].get('type_id2name', None)) and
            # UMLS has TypeIds that start with T
            # so in this case we can be _pretty sure_ it's Snomed
            not any("T" in tui for tui in all_tuis)):
        tid2name = {
            tid: name for tid, name in _DEFAULT_SNOMED_TYPE_ID2NAME.items()
            if tid in all_tuis
        }
        print("Set TypeID2name for", len(tid2name),
              'TypeIDs out of', len(all_tuis))
    print("Conditions", bool(not tid2name),
          not any("T" in tui for tui in all_tuis),
          "IN", all_tuis)
    print("Got TypeID2name for", len(tid2name),
          'TypeIDs out of', len(all_tuis))
    for cui, tui in all_cui_tuis:
        if tui not in cdb.type_id2info:
            cdb.type_id2info[tui] = TypeInfo(
                type_id=tui, name=tid2name.get(tui), cuis=set())
        cdb.type_id2info[tui].cuis.add(cui)
    cdb.addl_info.update(data['addl_info'])
    # for cui, ontos in data['addl_info']['cui2ontologies'].items():
    #     cdb.cui2info[cui]['in_other_ontology'] = ontos
    return cdb


def _add_name_info(cdb: CDB, data: dict) -> CDB:
    all_names = set()
    for key in NAME2KEYS:
        if key in OPTIONAL_NAME2_KEYS and key not in data:
            continue
        cnames = data[key].keys()
        logger.debug("Adding %d names based on '%s", len(cnames), key)
        all_names.update(cnames)
    logger.info("A total of %d names found", len(all_names))
    logger.info("Adding names from cui2names")
    # add from cui2names
    for cui_infos in cdb.cui2info.values():
        all_names.update(cui_infos['names'])
    logger.info("A total of %d names found after adding from cui2names",
                len(all_names))
    # NOTE: name2cuis has the same cuis as name2cuis2status
    #       so v2 only uses the latter since it provides extra information
    name2cuis2status = data['name2cuis2status']
    name2cnt_train = data['name2count_train']
    name2is_upper = data.get('name_isupper', {})
    for name in all_names:
        cuis2status: dict[str, str] = {}
        _cuis2status = name2cuis2status.get(name, {})
        cuis2status.update(_cuis2status)
        cnt_train = name2cnt_train.get(name, 0)
        is_upper = name2is_upper.get(name, False)
        info = get_new_name_info(name, per_cui_status=cuis2status,
                                 is_upper=is_upper, count_train=cnt_train)
        cdb.name2info[name] = info
    return cdb


def update_names(cdb: CDB, data: dict):
    for name_from, name_to in TO_RENAME.items():
        setattr(cdb, name_to, data[name_from])


def convert_data(all_data: dict) -> CDB:
    """Convert the raw v1 data into a CDB.

    Args:
        all_data (dict): The raw v1 data off disk.

    Returns:
        CDB: The v2 CDB.
    """
    data = all_data['cdb']
    cdb = CDB(Config())
    cdb = _add_cui_info(cdb, data)
    cdb = _add_name_info(cdb, data)
    update_names(cdb, data)
    if 'config' in all_data:
        logger.info("Loading old style CDB with config included.")
        cdb.config = get_config_from_nested_dict(all_data['config'])
    return cdb


def get_cdb_from_old(old_path: str) -> CDB:
    """Get the v2 CDB from a v1 CDB path.

    Args:
        old_path (str): The v1 CDB path.

    Returns:
        CDB: The v2 CDB.
    """
    data = load_old_raw_data(old_path)
    return convert_data(data)



from typing import Optional, Union, Any, Iterator, cast
import random
import logging
from itertools import chain
from collections.abc import MutableMapping, KeysView
from contextlib import contextmanager

from medcat.cdb.cdb import CDB
from medcat.vocab import Vocab
from medcat.config.config import Config, SerialisableBaseModel, ComponentConfig
from medcat.utils.defaults import StatusTypes as ST
from medcat.utils.matutils import sigmoid
from medcat.utils.config_utils import temp_changed_config
from medcat.cdb.concepts import CUIInfo, TypeInfo, get_new_cui_info
from medcat.components.types import CoreComponentType, AbstractCoreComponent
from medcat.tokenizing.tokenizers import BaseTokenizer
from medcat.tokenizing.tokens import MutableEntity, MutableDocument
from medcat.components.linking.context_based_linker import (
    Linker as NormalLinker, PerDocumentTokenCache)
from medcat.components.linking.vector_context_model import ContextModel


logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())


TYPE_ID_PREFIX: str = "TYPE_ID:"


def add_tuis_to_cui_info(cui2info: dict[str, CUIInfo],
                         type_ids: dict[str, TypeInfo]
                         ):
    for tid, tid_info in type_ids.items():
        prefixed_tid = f"{TYPE_ID_PREFIX}{tid}"
        if prefixed_tid not in cui2info:
            cui2info[prefixed_tid] = get_new_cui_info(
                tid, preferred_name=tid_info.name, names={tid_info.name})


class TwoStepLinker(AbstractCoreComponent):
    """Link to a biomedical database.

    Args:
        cdb (CDB): The Context Database.
        vocab (Vocab): The vocabulary.
        config (Config): The config.
    """
    # Custom pipeline component name
    name = 'medcat2_two_step_linker'

    # Override
    def __init__(self, cdb: CDB, vocab: Vocab, config: Config) -> None:
        self.cdb = cdb
        self.vocab = vocab
        self.config = config
        self._init_cnf()
        self._linker = NormalLinker(cdb, vocab, config)
        self._linker.context_model._disamb_preprocessors.append(
            self._preprocess_disamb)
        add_tuis_to_cui_info(self.cdb.cui2info, self.cdb.type_id2info)
        self._tui_context_model = ContextModel(
            self.cdb.cui2info,
            self.cdb.name2info,
            self.cdb.weighted_average_function,
            self.vocab,
            self.config.components.linking,
            self.config.general.separator)

    def _init_cnf(self):
        if not isinstance(self.config.components.linking.additional,
                          TwoStepLinkerConfig):
            logger.info("Setting default TwoStepLinkerConfig instead of %s",
                        self.config.components.linking.additional)
            self.config.components.linking.additional = TwoStepLinkerConfig()

    def get_type(self) -> CoreComponentType:
        return CoreComponentType.linking

    def _train_tuis(self, tui: str, entity: MutableEntity,
                    doc: MutableDocument,
                    per_doc_valid_token_cache: PerDocumentTokenCache,
                    add_negative: bool = True) -> None:
        self._tui_context_model.train(
            tui, entity, doc, per_doc_valid_token_cache, negative=False)
        if (add_negative and
                self.config.components.linking.negative_probability
                >= random.random()):
            self._tui_context_model.train_using_negative_sampling(tui)

    def _process_entity_train_tuis(
            self, doc: MutableDocument, entity: MutableEntity,
            per_doc_valid_token_cache: PerDocumentTokenCache) -> None:
        # Check does it have a detected name
        name = entity.detected_name
        if name is None:
            return
        cuis = entity.link_candidates
        # ignore duplicates, but create a list
        per_cui_tuis = {cui: [
            f"{TYPE_ID_PREFIX}{tui}"
            for tui in self.cdb.cui2info[cui]['type_ids']] for cui in cuis}
        tuis = list(set(chain(*per_cui_tuis.values())))

        with changed_learning_rate(self.config, self.two_step_config):
            self._do_training(doc, entity, name, cuis,
                              per_cui_tuis, tuis, per_doc_valid_token_cache)

    def _do_training(self,
                     doc: MutableDocument,
                     entity: MutableEntity,
                     name: str,
                     cuis: list[str],
                     per_cui_tuis: dict[str, list[str]],
                     tuis: list[str],
                     per_doc_valid_token_cache: PerDocumentTokenCache):
        if len(tuis) == 1:
            self._train_tuis(
                tui=tuis[0], entity=entity, doc=doc,
                per_doc_valid_token_cache=per_doc_valid_token_cache)
        else:
            # TODO: find most appropriate CUIs
            name_info = self.cdb.name2info.get(name, None)
            if name_info is None:
                return
            for cui in cuis:
                if name_info['per_cui_status'][cui] not in ST.PRIMARY_STATUS:
                    continue
                # only primary names
                for tui in per_cui_tuis[cui]:
                    self._train_tuis(
                        tui=tui, entity=entity, doc=doc,
                        per_doc_valid_token_cache=per_doc_valid_token_cache)

    def _train_for_tuis(self, doc: MutableDocument) -> None:
        # Run training
        for entity in doc.ner_ents:
            self._process_entity_train_tuis(
                doc, entity, PerDocumentTokenCache())

    def _check_similarity(self, cui: str, context_similarity: float) -> bool:
        th_type = self.config.components.linking.similarity_threshold_type
        threshold = self.config.components.linking.similarity_threshold
        if th_type == 'static':
            return context_similarity >= threshold
        if th_type == 'dynamic':
            conf = self.cdb.cui2info[cui]['average_confidence']
            return context_similarity >= conf * threshold
        return False

    def _reweigh_candidates(
            self, doc: MutableDocument,
            entity: MutableEntity,
            per_doc_valid_token_cache: PerDocumentTokenCache,
            per_entity_weights: 'PerEntityWeights',
            ) -> None:
        # Check does it have a detected concepts
        cuis = entity.link_candidates
        if not cuis:
            return
        per_tui_candidates: dict[str, list[str]] = {}
        for cui in cuis:
            for tid in self.cdb.cui2info[cui]['type_ids']:
                if tid not in per_tui_candidates:
                    per_tui_candidates[tid] = []
                per_tui_candidates[tid].append(cui)
        # NOTE: adding prefix to differentiate from regular CUIs
        tuis = [f"{TYPE_ID_PREFIX}{tid}" for tid in per_tui_candidates]
        # TODO: how do I only use ones that are allowed filter-wise?
        cnf_l = self.config.components.linking
        if cnf_l.filters.cuis:
            # this should check whether a TUI that's listed above corresponds
            # to any CUIs in the list of allowed ones
            # and only keep the ones that do
            num_before = len(tuis)
            allowed_tuis = set(
                f"{TYPE_ID_PREFIX}{tui4cui}" for cui in cnf_l.filters.cuis
                if cui in self.cdb.cui2info
                for tui4cui in self.cdb.cui2info[cui]['type_ids'])
            tuis = list(filter(allowed_tuis.__contains__, tuis))
            logger.debug("Filtered from %d to %d due to %d CUIs in filters",
                         num_before, len(tuis), len(cnf_l.filters.cuis))
        if cnf_l.filters.cuis_exclude:
            # this should check whether a TUI that's listed above corresponds
            # to any CUIs in the list of excluded ones
            # and only keep ones that don't
            raise ValueError(
                "TODO - filter based on EXCLUDED cuis: "
                f"{cnf_l.filters.cuis_exclude}")
        name = entity.detected_name
        if name is None:
            # ignore if there's no name found
            # in my experience, this shouldn't really happen anyway
            return
        (suitable_tuis,
         tui_similarities, _) = self._tui_context_model.get_all_similarities(
            tuis, entity, name, doc, per_doc_valid_token_cache)
        per_cui_type_sims = {
            cui: sim
            for tui, sim in zip(suitable_tuis, tui_similarities)
            if tui is not None
            for cui in per_tui_candidates[tui.removeprefix(TYPE_ID_PREFIX)]
        }
        per_entity_weights[entity] = per_cui_type_sims
        logger.debug("Adding per CUI to %s (tokens %d..%d) weights %s",
                     cui, entity.base.start_index, entity.base.end_index,
                     per_cui_type_sims)

    def _weigh_on_inference(self, doc: MutableDocument) -> 'PerEntityWeights':
        per_entity_weights = PerEntityWeights(doc)
        per_doc_valid_token_cache = PerDocumentTokenCache()
        for entity in doc.ner_ents:
            logger.debug("Narrowing down candidates for: '%s' from %s",
                         entity.base.text, entity.link_candidates)
            self._reweigh_candidates(
                doc, entity, per_doc_valid_token_cache, per_entity_weights)
        return per_entity_weights

    def __call__(self, doc: MutableDocument) -> MutableDocument:
        per_ent_weights: Optional[PerEntityWeights] = None
        if self.config.components.linking.train:
            self._train_for_tuis(doc)
        else:
            per_ent_weights = self._weigh_on_inference(doc)
        with temp_attribute(self, '_per_ent_weights', per_ent_weights):
            return self._linker(doc)

    def train(self, cui: str,
              entity: MutableEntity,
              doc: MutableDocument,
              negative: bool = False,
              names: Union[list[str], dict] = []) -> None:
        """Train the linker.

        This simply trains the context model.

        Args:
            cui (str): The CUI to train.
            entity (BaseEntity): The entity we're at.
            doc (BaseDocument): The document within which we're working.
            negative (bool): Whether or not the example is negative.
                Defaults to False.
            names (list[str]/dict):
                Optionally used to update the `status` of a name-cui
                pair in the CDB.
        """
        pdc = PerDocumentTokenCache()
        tuis = self.cdb.cui2info[cui]['type_ids']
        for tui in tuis:
            # one CUI may have multiple type IDs
            tui = next(iter(tuis))
            self._tui_context_model.train(f"{TYPE_ID_PREFIX}{tui}",
                                          entity, doc, pdc,
                                          negative=negative, names=names)
        self._linker.train(cui, entity, doc, negative, names,
                           per_doc_valid_token_cache=pdc)

    @classmethod
    def create_new_component(
            cls, cnf: ComponentConfig, tokenizer: BaseTokenizer,
            cdb: CDB, vocab: Vocab, model_load_path: Optional[str]
            ) -> 'TwoStepLinker':
        return cls(cdb, vocab, cdb.config)

    @property
    def two_step_config(self) -> 'TwoStepLinkerConfig':
        cnf_2step = self.config.components.linking.additional
        if not isinstance(cnf_2step, TwoStepLinkerConfig):
            raise ValueError(
                "Need to set 'config.components.linking.additional' to "
                "an instance of TwoStepLinkerConfig")
        return cnf_2step

    _DEBUG_CNT = 0  # TODO: remove

    def _preprocess_disamb(self, ent: MutableEntity, name: str,
                           cuis: list[str], similarities: list[float]) -> None:
        if cuis and all(cui.startswith(TYPE_ID_PREFIX) for cui in cuis):
            return
        if not hasattr(self, '_per_ent_weights'):
            raise ValueError("No per entity weights")
        pew = self._per_ent_weights
        if ent not in pew:
            # ignore for stuff not in here
            return
        per_cui_type_sims = pew[ent]
        cnf_2step = self.two_step_config
        for cui, type_sim in per_cui_type_sims.items():
            if cui not in cuis:
                continue
            cui_index = cuis.index(cui)
            cui_sim = similarities[cui_index]
            ts_coef = sigmoid(
                cnf_2step.alpha_sharpness * (
                    type_sim - cnf_2step.alpha_midpoint))
            logger.debug(
                "Mixing type similarity of %.4f and CUI similarity of %.4f "
                "with %.4f weight for CUI similarity",
                cui_sim, type_sim, ts_coef)
            new_sim = (1 - ts_coef) * cui_sim + ts_coef * type_sim
            similarities[cui_index] = new_sim
            logger.debug("[Per CUI weights] CUI: %s, Name: %s, "
                         "Old sim: %.3f, New sim: %.3f",
                         cui, name, cui_sim, new_sim)


class PerEntityWeights(MutableMapping[MutableEntity, dict[str, float]]):

    def __init__(self, doc: MutableDocument):
        self._doc = doc
        self._cui_weights: dict[tuple[int, int], dict[str, float]] = {}

    def _to_key(self, ent: MutableEntity) -> tuple[int, int]:
        return ent.base.start_index, ent.base.end_index

    def _from_key(self, key: tuple[int, int]) -> MutableEntity:
        start, end = key
        for ent in self._doc.ner_ents:
            if ent.base.start_index == start and ent.base.end_index == end:
                return ent
        raise ValueError("Unable to find entity corresponding to "
                         f"{start}...{end} (token indices) within doc "
                         f"{self._doc}")

    def __getitem__(self, ent: MutableEntity):
        return self._cui_weights[self._to_key(ent)]

    def __contains__(self, key: object):
        # NOTE: shouldn't ever really be anything else
        ent = cast(MutableEntity, key)
        return self._to_key(ent) in self._cui_weights

    def __setitem__(self, ent: MutableEntity, value: dict[str, float]):
        self._cui_weights[self._to_key(ent)] = value

    def __delitem__(self, key: MutableEntity) -> None:
        del self._cui_weights[self._to_key(key)]

    def __iter__(self) -> Iterator[MutableEntity]:
        return (self._from_key(k) for k in self._cui_weights)

    def __len__(self) -> int:
        return len(self._cui_weights)

    def keys(self) -> KeysView[MutableEntity]:
        return {self._from_key(k): None for k in self._cui_weights}.keys()


def changed_learning_rate(config: Config, two_step_cnf: 'TwoStepLinkerConfig'):
    coef = two_step_cnf.type_learning_rate_coefficient
    comp_optim = config.components.linking.optim.copy()
    for learning_rate_name in comp_optim:
        if 'lr' not in learning_rate_name:
            continue
        comp_optim[learning_rate_name] *= coef
    logger.debug("Changing learning rate from %s to %s",
                 config.components.linking.optim,
                 comp_optim)
    return temp_changed_config(
        config.components.linking, 'optim', comp_optim)


class TwoStepLinkerConfig(SerialisableBaseModel):
    alpha_midpoint: float = 0.5
    """The midpoint for the sigmoid.
    alpha = sigmoid(alpha_sharpness(similarity - alpha_midpoint))
    This is used for weighting the type similarity vs the concept similarity.
    """
    alpha_sharpness: float = 5.0
    """The sharpness for the sigmoid.
    alpha = sigmoid(alpha_sharpness(similarity - alpha_midpoint))
    This is used for weighting the type similarity vs the concept similarity.
    """
    type_learning_rate_coefficient: float = 0.2
    """The coefficient for the type-based context model learning rate.

    The idea is that since there's a far fewer classes for types,
    we need to lower the learning rate. In the Snomed examples we have
    around 10 000 more CUIs then types so a coefficient like 0.2 should be
    appropriate.
    """


@contextmanager
def temp_attribute(obj: Any, attr_name: str, attr_val: Any):
    if hasattr(obj, attr_name):
        prev_val = getattr(obj, attr_name)
        logger.warning(
            "Object '%s' already had an attribute '%s' - has type '%s'",
            obj, attr_val, type(prev_val))
    else:
        prev_val = None
    setattr(obj, attr_name, attr_val)
    yield
    # and reset
    if prev_val is not None:
        setattr(obj, attr_name, prev_val)
    else:
        delattr(obj, attr_name)



from typing import Optional, TypedDict
from dataclasses import dataclass, field

import numpy as np


class CUIInfo(TypedDict):
    cui: str  # NOTE: we _could_ get away without to save on memory
    preferred_name: str
    names: set[str]
    subnames: set[str]
    type_ids: set[str]
    # optional parts start here
    description: Optional[str]
    original_names: Optional[set[str]]
    tags: Optional[list[str]]
    group: Optional[str]
    in_other_ontology: Optional[set[str]]
    # stuff related to training starts here
    # TODO: separate supervised and unsupervised
    count_train: int
    context_vectors: Optional[dict[str, np.ndarray]]
    average_confidence: float


def get_new_cui_info(cui: str, preferred_name: str,
                     names: set[str] = set(),
                     subnames: set[str] = set(),
                     type_ids: set[str] = set(),
                     description: Optional[str] = None,
                     original_names: Optional[set[str]] = None,
                     tags: Optional[list[str]] = None,
                     group: Optional[str] = None,
                     in_other_ontology: Optional[set[str]] = None,
                     count_train: int = 0,
                     context_vectors: Optional[dict[str, np.ndarray]] = None,
                     average_confidence: float = 0.0) -> CUIInfo:
    return {
        'cui': cui,
        'preferred_name': preferred_name,
        'names': names or names.copy(),
        'subnames': subnames or subnames.copy(),
        'type_ids': type_ids or type_ids.copy(),
        'description': description,
        'original_names': original_names,
        'tags': tags,
        'group': group,
        'in_other_ontology': in_other_ontology,
        'count_train': count_train,
        'context_vectors': context_vectors,
        'average_confidence': average_confidence
    }


def reset_cui_training(cui_info: CUIInfo) -> None:
    cui_info['context_vectors'] = None
    cui_info['count_train'] = 0
    cui_info['average_confidence'] = 0


class NameInfo(TypedDict):
    name: str  # NOTE: we _could_ get away without to save on memory
    per_cui_status: dict[str, str]
    is_upper: bool
    # stuff related to training starts here
    count_train: int


def get_new_name_info(name: str,
                      per_cui_status: dict[str, str] = {},
                      is_upper: bool = False,
                      count_train: int = 0) -> NameInfo:
    return {
        'name': name,
        'per_cui_status': per_cui_status or per_cui_status.copy(),
        'is_upper': is_upper,
        'count_train': count_train
    }


@dataclass
class TypeInfo:
    """Represents all the info regarding a type ID."""
    type_id: str  # NOTE: we _could_ get away without to save on memory
    name: str
    cuis: set[str] = field(default_factory=set)



import os
from transformers import PreTrainedTokenizerFast
import logging

from medcat.config.config_rel_cat import ConfigRelCAT
from medcat.components.addons.relation_extraction.tokenizer import (
    BaseTokenizerWrapper)

logger = logging.getLogger(__name__)


class TokenizerWrapperModernBERT(BaseTokenizerWrapper):
    """Wrapper around a huggingface ModernBERT tokenizer so that it works
    with the RelCAT models.

    Args:
        hf_tokenizers (`transformers.PreTrainedTokenizerFast`):
            A huggingface Fast tokenizer.
    """
    name = "tokenizer_wrapper_modern_bert_rel"
    pretrained_model_name_or_path = "answerdotai/ModernBERT-base"

    @classmethod
    def load(cls, tokenizer_path: str, relcat_config: ConfigRelCAT, **kwargs
             ) -> "TokenizerWrapperModernBERT":
        tokenizer = cls()
        path = os.path.join(tokenizer_path, cls.name)

        if tokenizer_path:
            tokenizer.hf_tokenizers = PreTrainedTokenizerFast.from_pretrained(
                path, **kwargs)
        else:
            relcat_config.general.model_name = (
                cls.pretrained_model_name_or_path)
            tokenizer.hf_tokenizers = PreTrainedTokenizerFast.from_pretrained(
                path=relcat_config.general.model_name)
        return tokenizer



from typing import TypedDict, Optional


class MetaAnnotation(TypedDict):
    value: str
    confidence: float
    name: str


class Entity(TypedDict, total=False):
    pretty_name: str
    cui: str
    type_ids: list[str]
    # types: list[str]# TODO: add back
    source_value: str
    detected_name: str
    acc: float
    context_similarity: float
    start: int
    end: int
    # icd10: list[str]# TODO: add back
    # ontologies: list[str]# TODO: add back
    # snomed: list[str]# TODO: add back
    id: int
    meta_anns: dict[str, MetaAnnotation]
    # optionals:
    context_left: list[str]
    context_center: list[str]
    context_right: list[str]


class Entities(TypedDict):
    # We use int as the key type, but a generic dictionary with int keys
    # is defined because the keys in entities are dynamic and can change.
    entities: dict[int, Entity]
    tokens: list[str]  # TODO - do we need this
    text: Optional[str]


class OnlyCUIEntities(TypedDict):
    entities: dict[int, str]
    tokens: list[str]  # TODO - do we need this
    text: Optional[str]
